# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'Pazar',
        1: 'Pazartesi',
        2: 'Salı',
        3: 'Çarşamba',
        4: 'Perşembe',
        5: 'Cuma',
        6: 'Cumartesi'
    },
    'days_abbrev': {
        0: 'Paz',
        1: 'Pzt',
        2: 'Sal',
        3: 'Çar',
        4: 'Per',
        5: 'Cum',
        6: 'Cmt'
    },

    # Months
    'months': {
        1: 'Ocak',
        2: 'Şubat',
        3: 'Mart',
        4: 'Nisan',
        5: 'Mayıs',
        6: 'Haziran',
        7: 'Temmuz',
        8: 'Ağustos',
        9: 'Eylül',
        10: 'Ekim',
        11: 'Kasım',
        12: 'Aralık',
    },
    'months_abbrev': {
        1: 'Oca',
        2: 'Şub',
        3: 'Mar',
        4: 'Nis',
        5: 'May',
        6: 'Haz',
        7: 'Tem',
        8: 'Ağu',
        9: 'Eyl',
        10: 'Eki',
        11: 'Kas',
        12: 'Ara',
    },

    # Units of time
    'year': '{count} yıl',
    'month': '{count} ay',
    'week': '{count} hafta',
    'day': '{count} gün',
    'hour': '{count} saat',
    'minute': '{count} dakika',
    'second': '{count} saniye',

    # Relative time
    'ago': '{time} önce',
    'from_now': '{time} sonra',
    'after': '{time} sonra',
    'before': '{time} önce',

    # Date formats
    'date_formats': {
        'LTS': 'HH:mm:ss',
        'LT': 'HH:mm',
        'LLLL': 'dddd, D MMMM YYYY HH:mm',
        'LLL': 'D MMMM YYYY HH:mm',
        'LL': 'D MMMM YYYY',
        'L': 'DD.MM.YYYY',
    },
}



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'svētdiena',
        1: 'pirmdiena',
        2: 'otrdiena',
        3: 'trešdiena',
        4: 'ceturtdiena',
        5: 'piektdiena',
        6: 'sestdiena'
    },
    'days_abbrev': {
        0: 'Sv',
        1: 'P',
        2: 'O',
        3: 'T',
        4: 'C',
        5: 'Pk',
        6: 'S'
    },

    # Months
    'months': {
        1: 'janvāris',
        2: 'februāris',
        3: 'marts',
        4: 'aprīlis',
        5: 'maijs',
        6: 'jūnijs',
        7: 'jūlijs',
        8: 'augusts',
        9: 'septembris',
        10: 'oktobris',
        11: 'novembris',
        12: 'decembris'
    },
    'months_abbrev': {
        1: 'jan',
        2: 'feb',
        3: 'mar',
        4: 'apr',
        5: 'mai',
        6: 'jūn',
        7: 'jūl',
        8: 'aug',
        9: 'sep',
        10: 'okt',
        11: 'nov',
        12: 'dec'
    },

    # Units of time
    'year': ['0 gadiem', '{count} gada', '{count} gadiem'],
    'month': ['0 mēnešiem', '{count} mēneša', '{count} mēnešiem'],
    'week': ['0 nedēļām', '{count} nedēļas', '{count} nedēļām'],
    'day': ['0 dienām', '{count} dienas', '{count} dienām'],
    'hour': ['0 stundām', '{count} stundas', '{count} stundām'],
    'minute': ['0 minūtēm', '{count} minūtes', '{count} minūtēm'],
    'second': ['0 sekundēm', '{count} sekundes', '{count} sekundēm'],

    # Relative time
    'ago': 'pirms {time}',
    'from_now': 'pēc {time}',
    'after': '{time} vēlāk',
    'before': '{time} pirms',

    'year_after': ['0 gadus', '{count} gadu', '{count} gadus'],
    'month_after': ['0 mēnešus', '{count} mēnesi', '{count} mēnešus'],
    'week_after': ['0 nedēļas', '{count} nedēļu', '{count} nedēļas'],
    'day_after': ['0 dienas', '{count} dienu', '{count} dienas'],
    'hour_after': ['0 stundas', '{count} stundu', '{count} stundas'],
    'minute_after': ['0 minūtes', '{count} minūti', '{count} minūtes'],
    'second_after': ['0 sekundes', '{count} sekundi', '{count} sekundes'],

    'year_before': ['0 gadus', '{count} gadu', '{count} gadus'],
    'month_before': ['0 mēnešus', '{count} mēnesi', '{count} mēnešus'],
    'week_before': ['0 nedēļas', '{count} nedēļu', '{count} nedēļas'],
    'day_before': ['0 dienas', '{count} dienu', '{count} dienas'],
    'hour_before': ['0 stundas', '{count} stundu', '{count} stundas'],
    'minute_before': ['0 minūtes', '{count} minūti', '{count} minūtes'],
    'second_before': ['0 sekundes', '{count} sekundi', '{count} sekundes'],

    # Date formats
    'date_formats': {
        'LTS': 'HH:mm:ss',
        'LT': 'HH:mm',
        'LLLL': 'YYYY. [gada] D. MMMM, dddd, HH:mm',
        'LLL': 'YYYY. [gada] D. MMMM, HH:mm',
        'LL': 'YYYY. [gada] D. MMMM',
        'L': 'DD.MM.YYYY.',
    },
}



# -*- coding: utf-8 -*-

import operator
import pendulum

from datetime import datetime, date, timedelta

from .mixins.interval import WordableIntervalMixin
from .interval import BaseInterval, Interval
from .constants import MONTHS_PER_YEAR
from .helpers import precise_diff


class Period(WordableIntervalMixin, BaseInterval):
    """
    Interval class that is aware of the datetimes that generated the
    time difference.
    """

    def __new__(cls, start, end, absolute=False):
        if absolute and start > end:
            end, start = start, end

        if isinstance(start, pendulum.Pendulum):
            start = start._datetime
        elif isinstance(start, pendulum.Date):
            start = date(start.year, start.month, start.day)

        if isinstance(end, pendulum.Pendulum):
            end = end._datetime
        elif isinstance(end, pendulum.Date):
            end = date(end.year, end.month, end.day)

        delta = end - start

        return super(Period, cls).__new__(
            cls, seconds=delta.total_seconds()
        )

    def __init__(self, start, end, absolute=False):
        super(Period, self).__init__()

        if not isinstance(start, (pendulum.Date)):
            if isinstance(start, datetime):
                start = pendulum.Pendulum.instance(start)
            else:
                start = pendulum.Date.instance(start)

            _start = start
        else:
            if isinstance(start, pendulum.Pendulum):
                _start = start._datetime
            else:
                _start = date(start.year, start.month, start.day)

        if not isinstance(end, (pendulum.Date)):
            if isinstance(end, datetime):
                end = pendulum.Pendulum.instance(end)
            else:
                end = pendulum.Date.instance(end)

            _end = end
        else:
            if isinstance(end, pendulum.Pendulum):
                _end = end._datetime
            else:
                _end = date(end.year, end.month, end.day)

        self._invert = False
        if start > end:
            self._invert = True

            if absolute:
                end, start = start, end
                _end, _start = _start, _end

        self._absolute = absolute
        self._start = start
        self._end = end
        self._delta = precise_diff(_start, _end)

    @property
    def years(self):
        return self._delta['years']

    @property
    def months(self):
        return self._delta['months']

    @property
    def weeks(self):
        return self._delta['days'] // 7

    @property
    def days(self):
        return self._days

    @property
    def remaining_days(self):
        return abs(self._delta['days']) % 7 * self._sign(self._days)

    @property
    def hours(self):
        return self._delta['hours']

    @property
    def minutes(self):
        return self._delta['minutes']

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end

    def in_years(self):
        """
        Gives the duration of the Period in full years.

        :rtype: int
        """
        return self.years

    def in_months(self):
        """
        Gives the duration of the Period in full months.

        :rtype: int
        """
        return self.years * MONTHS_PER_YEAR + self.months

    def in_weeks(self):
        days = self.in_days()
        sign = 1

        if days < 0:
            sign = -1

        return sign * (abs(days) // 7)

    def in_days(self):
        return self._delta['total']['days']

    def in_weekdays(self):
        start, end = self.start.start_of('day'), self.end.start_of('day')
        if not self._absolute and self.invert:
            start, end = self.end.start_of('day'), self.start.start_of('day')

        days = 0
        while start <= end:
            if start.is_weekday():
                days += 1

            start = start.add(days=1)

        return days * (-1 if not self._absolute and self.invert else 1)

    def in_weekend_days(self):
        start, end = self.start.start_of('day'), self.end.start_of('day')
        if not self._absolute and self.invert:
            start, end = self.end.start_of('day'), self.start.start_of('day')

        days = 0
        while start <= end:
            if start.is_weekend():
                days += 1

            start = start.add(days=1)

        return days * (-1 if not self._absolute and self.invert else 1)

    def in_words(self, locale=None, separator=' '):
        """
        Get the current interval in words in the current locale.

        Ex: 6 jours 23 heures 58 minutes

        :param locale: The locale to use. Defaults to current locale.
        :type locale: str

        :param separator: The separator to use between each unit
        :type separator: str

        :rtype: str
        """
        periods = [
            ('year', self.years),
            ('month', self.months),
            ('week', self.weeks),
            ('day', self.remaining_days),
            ('hour', self.hours),
            ('minute', self.minutes),
            ('second', self.remaining_seconds)
        ]

        return super(Period, self).in_words(
            locale=locale, separator=separator, _periods=periods
        )

    def range(self, unit, amount=1):
        return list(self.xrange(unit, amount))

    def xrange(self, unit, amount=1):
        method = 'add'
        op = operator.le
        if not self._absolute and self.invert:
            method = 'subtract'
            op = operator.ge

        start, end = self.start, self.end

        i = amount
        while op(start, end):
            yield start

            start = getattr(self.start, method)(**{unit: i})

            i += amount

    def intersect(self, *periods):
        """
        Return the Period intersection of the current Period
        and the given periods.

        :type periods: tuple of Period

        :rtype: Period
        """
        start, end = self.start, self.end
        has_intersection = False
        for period in periods:
            if period.end < start or period.start > end:
                continue

            has_intersection = True
            start = max(start, period.start)
            end = min(end, period.end)

        if not has_intersection:
            return None

        return self.__class__(start, end)

    def as_interval(self):
        """
        Return the Period as an Interval.

        :rtype: Interval
        """
        return Interval(seconds=self.total_seconds())

    def __iter__(self):
        return self.xrange('days')

    def __contains__(self, item):
        from .pendulum import Pendulum

        if not isinstance(item, Pendulum):
            item = Pendulum.instance(item)

        return item.between(self.start, self.end)

    def __add__(self, other):
        return self.as_interval().__add__(other)

    __radd__ = __add__

    def __sub__(self, other):
        return self.as_interval().__sub__(other)

    def __neg__(self):
        return self.__class__(self.end, self.start, self._absolute)

    def __mul__(self, other):
        return self.as_interval().__mul__(other)

    __rmul__ = __mul__

    def __floordiv__(self, other):
        return self.as_interval().__floordiv__(other)

    def __truediv__(self, other):
        return self.as_interval().__truediv__(other)

    __div__ = __floordiv__

    def __mod__(self, other):
        return self.as_interval().__mod__(other)

    def __divmod__(self, other):
        return self.as_interval().__divmod__(other)

    def __abs__(self):
        return self.__class__(self.start, self.end, True)

    def __repr__(self):
        return '<Period [{} -> {}]>'.format(
            self._start, self._end
        )

    def _cmp(self, other):
        # Only needed for PyPy3
        assert isinstance(other, timedelta)

        if isinstance(other, Period):
            other = other.as_timedelta()

        td = self.as_timedelta()

        return 0 if td == other else 1 if td > other else -1

    def _getstate(self, protocol=3):
        start, end = self.start, self.end

        if self._invert and self._absolute:
            end, start = start, end

        return (
            start, end, self._absolute
        )

    def __reduce__(self):
        return self.__reduce_ex__(2)

    def __reduce_ex__(self, protocol):
        return self.__class__, self._getstate(protocol)



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'Якшанба',
        1: 'Душанба',
        2: 'Сешанба',
        3: 'Чоршанба',
        4: 'Пайшанба',
        5: 'Жума',
        6: 'Шанба'
    },
    'days_abbrev': {
        0: 'Якш',
        1: 'Душ',
        2: 'Сеш',
        3: 'Чор',
        4: 'Пай',
        5: 'Жум',
        6: 'Шан'
    },

    # Months
    'months': {
        1: 'январ',
        2: 'феврал',
        3: 'март',
        4: 'апрел',
        5: 'май',
        6: 'июн',
        7: 'июл',
        8: 'август',
        9: 'сентябр',
        10: 'октябр',
        11: 'ноябр',
        12: 'декабр'
    },
    'months_abbrev': {
        1: 'янв',
        2: 'фев',
        3: 'мар',
        4: 'апр',
        5: 'май',
        6: 'июн',
        7: 'июл',
        8: 'авг',
        9: 'сен',
        10: 'окт',
        11: 'ноя',
        12: 'дек'
    },

    # Units of time
    'year': ['{count} yil', '{count} yil', '{count} yil'],
    'month': ['{count} oy', '{count} oy', '{count} oylar'],
    'week': ['{count} hafta', '{count} hafta', '{count} hafta'],
    'day': ['{count} kun', '{count} kun', '{count} kun'],
    'hour': ['{count} soat', '{count} soat', '{count} soat'],
    'minute': ['{count} minut', '{count} minut', '{count} minut'],
    'second': ['{count} sekund', '{count} sekund', '{count} sekund'],

    # Relative time
    'ago': '{time} avval',
    'from_now': 'keyin {time}',
    'after': '{time} keyin',
    'before': '{time} gacha',

    # Date formats
    'date_formats': {
        'LTS': 'HH:mm:ss',
        'LT': 'HH:mm',
        'LLLL': 'D MMMM YYYY, dddd HH:mm',
        'LLL': 'D MMMM YYYY HH:mm',
        'LL': 'D MMMM YYYY',
        'L': 'DD/MM/YYYY',
    },
}



# -*- coding: utf-8 -*-

# Types
from .pendulum import Pendulum
from .date import Date
from .time import Time
from .interval import Interval
from .period import Period

# Mimicking standard library
datetime = Pendulum
date = Date
time = Time

# Constants
from .constants import (
    MONDAY, TUESDAY, WEDNESDAY,
    THURSDAY, FRIDAY, SATURDAY, SUNDAY,
    YEARS_PER_CENTURY, YEARS_PER_DECADE,
    MONTHS_PER_YEAR, WEEKS_PER_YEAR, DAYS_PER_WEEK,
    HOURS_PER_DAY, MINUTES_PER_HOUR, SECONDS_PER_MINUTE,
    SECONDS_PER_HOUR, SECONDS_PER_DAY
)

from .tz.timezone import Timezone

from .mixins.default import TranslatableMixin, FormattableMixing, TestableMixin

from ._global import Global

PRE_TRANSITION = Timezone.PRE_TRANSITION
POST_TRANSITION = Timezone.POST_TRANSITION
TRANSITION_ERROR = Timezone.TRANSITION_ERROR

# Global helpers
test = Global.test
set_test_now = Global.set_test_now
has_test_now = Global.has_test_now
get_test_now = Global.get_test_now
set_locale = Global.set_locale
get_locale = Global.get_locale
translator = Global.translator
set_translator = Global.set_translator
set_formatter = Global.set_formatter
get_formatter = Global.get_formatter

# Helpers
from .parser import parse

instance = Pendulum.instance
now = Pendulum.now
utcnow = Pendulum.utcnow
today = Pendulum.today
tomorrow = Pendulum.tomorrow
yesterday = Pendulum.yesterday
create = Pendulum.create
from_format = Pendulum.create_from_format
strptime = Pendulum.strptime
from_timestamp = Pendulum.create_from_timestamp
set_to_string_format = Pendulum.set_to_string_format
reset_to_string_format = Pendulum.reset_to_string_format
set_transition_rule = Pendulum.set_transition_rule
get_transition_rule = Pendulum.get_transition_rule

# Standard helpers
min = Pendulum.min
max = Pendulum.max
fromtimestamp = Pendulum.fromtimestamp
utcfromtimestamp = Pendulum.utcfromtimestamp
fromordinal = Pendulum.fromordinal
combine = Pendulum.combine

# Interval
interval = Interval

# Period
period = Period

# Timezones
from .tz import timezone, timezones, local_timezone, UTC



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'niedziela',
        1: 'poniedziałek',
        2: 'wtorek',
        3: 'środa',
        4: 'czwartek',
        5: 'piątek',
        6: 'sobota'
    },
    'days_abbrev': {
        0: 'Nd',
        1: 'Pn',
        2: 'Wt',
        3: 'Śr',
        4: 'Czw',
        5: 'Pt',
        6: 'So'
    },

    # Months
    'months': {
        1: 'styczeń',
        2: 'luty',
        3: 'marzec',
        4: 'kwiecień',
        5: 'maj',
        6: 'czerwiec',
        7: 'lipiec',
        8: 'sierpień',
        9: 'wrzesień',
        10: 'październik',
        11: 'listopad',
        12: 'grudzień',
    },
    'months_abbrev': {
        1: 'sty',
        2: 'lut',
        3: 'mar',
        4: 'kwi',
        5: 'maj',
        6: 'cze',
        7: 'lip',
        8: 'sie',
        9: 'wrz',
        10: 'paź',
        11: 'lis',
        12: 'gru',
    },

    # Units of time
    'year': ['{count} rok', '{count} lata', '{count} lat'],
    'month': ['{count} miesiąc', '{count} miesiące', '{count} miesięcy'],
    'week': ['{count} tydzień', '{count} tygodnie', '{count} tygodni'],
    'day': ['{count} dzień', '{count} dni', '{count} dni'],
    'hour': ['{count} godzina', '{count} godziny', '{count} godzin'],
    'minute': ['{count} minuta', '{count} minuty', '{count} minut'],
    'second': ['{count} sekunda', '{count} sekundy', '{count} sekund'],

    # Relative time
    'ago': '{time} temu',
    'from_now': '{time} od teraz',
    'after': '{time} po',
    'before': '{time} przed',

    # Date formats
    'date_formats': {
        'LTS': 'HH:mm:ss',
        'LT': 'HH:mm',
        'LLLL': 'dddd, D MMMM YYYY HH:mm',
        'LLL': 'D MMMM YYYY HH:mm',
        'LL': 'D MMMM YYYY',
        'L': 'DD.MM.YYYY',
    },
}



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'neděle',
        1: 'pondělí',
        2: 'úterý',
        3: 'středa',
        4: 'čtvrtek',
        5: 'pátek',
        6: 'sobota'
    },
    'days_abbrev': {
        0: 'ne',
        1: 'po',
        2: 'út',
        3: 'st',
        4: 'čt',
        5: 'pá',
        6: 'so'
    },

    # Months
    'months': {
        1: 'leden',
        2: 'únor',
        3: 'březen',
        4: 'duben',
        5: 'květen',
        6: 'červen',
        7: 'červenec',
        8: 'srpen',
        9: 'září',
        10: 'říjen',
        11: 'listopad',
        12: 'prosinec',
    },
    'months_abbrev': {
        1: 'led',
        2: 'úno',
        3: 'bře',
        4: 'dub',
        5: 'kvě',
        6: 'čvn',
        7: 'čvc',
        8: 'srp',
        9: 'zář',
        10: 'říj',
        11: 'lis',
        12: 'pro',
    },

    # Units of time
    'year': ['{count} rok', '{count} roky', '{count} let'],
    'month': ['{count} měsíc', '{count} měsíce', '{count} měsíců'],
    'week': ['{count} týden', '{count} týdny', '{count} týdnů'],
    'day': ['{count} den', '{count} dny', '{count} dní'],
    'hour': ['{count} hodina', '{count} hodiny', '{count} hodin'],
    'minute': ['{count} minuta', '{count} minuty', '{count} minut'],
    'second': ['{count} sekunda', '{count} sekundy', '{count} sekund'],

    # Relative time
    'ago': '{time} nazpět',
    'from_now': 'za {time}',
    'after': '{time} později',
    'before': '{time} předtím',

    # Date formats
    'date_formats': {
        'LTS': 'H:mm:ss',
        'LT': 'H:mm',
        'LLLL': 'dddd D. MMMM YYYY H:mm',
        'LLL': 'D. MMMM YYYY H:mm',
        'LL': 'D. MMMM YYYY',
        'L': 'DD.MM.YYYY',
    },
}



# -*- coding: utf-8 -*-

import re
import datetime

from .._compat import decode
from .formatter import Formatter


_MATCH_1 = re.compile('\d')
_MATCH_2 = re.compile('\d\d')
_MATCH_3 = re.compile('\d{3}')
_MATCH_4 = re.compile('\d{4}')
_MATCH_6 = re.compile('[+-]?\d{6}')
_MATCH_1_TO_2 = re.compile('\d\d?')
_MATCH_1_TO_3 = re.compile('\d{1,3}')
_MATCH_1_TO_4 = re.compile('\d{1,4}')
_MATCH_1_TO_6 = re.compile('[+-]?\d{1,6}')
_MATCH_3_TO_4 = re.compile('\d{3}\d?')
_MATCH_5_TO_6 = re.compile('\d{5}\d?')
_MATCH_UNSIGNED = re.compile('\d+')
_MATCH_SIGNED = re.compile('[+-]?\d+')
_MATCH_OFFSET = re.compile('(?i)Z|[+-]\d\d:?\d\d')
_MATCH_SHORT_OFFSET = re.compile('(?i)Z|[+-]\d\d(?::?\d\d)?')
_MATCH_TIMESTAMP = re.compile('[+-]?\d+(\.\d{1,3})?')
_MATCH_WORD = re.compile("[0-9]*['a-z\u00A0-\u05FF\u0700-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]+|[\u0600-\u06FF\/]+(\s*?[\u0600-\u06FF]+){1,2}")



class AlternativeFormatter(Formatter):

    _TOKENS = '\[([^\[]*)\]|\\\(.)|' \
              '(' \
              'Mo|MM?M?M?' \
              '|Do|DDDo|DD?D?D?|ddd?d?|do?' \
              '|w[o|w]?|W[o|W]?|Qo?' \
              '|YYYY|YY|Y' \
              '|gg(ggg?)?|GG(GGG?)?' \
              '|e|E|a|A' \
              '|hh?|HH?|kk?' \
              '|mm?|ss?|S{1,9}' \
              '|x|X' \
              '|zz?|ZZ?' \
              '|LTS|LT|LL?L?L?' \
              ')'

    _FORMAT_RE = re.compile(_TOKENS)

    _LOCALIZABLE_TOKENS = (
        'Mo', 'MMM', 'MMMM',
        'Qo',
        'Do',
        'DDDo',
        'do', 'dd', 'ddd', 'dddd',
        'wo',
        'Wo',
        'A', 'a',
    )

    _TOKENS_RULES = {
        # Year
        'YYYY': lambda dt: '{:d}'.format(dt.year),
        'YY': lambda dt: '{:d}'.format(dt.year)[2:],
        'Y': lambda dt: '{:d}'.format(dt.year),

        # Quarter
        'Q': lambda dt: '{:d}'.format(dt.quarter),

        # Month
        'MM': lambda dt: '{:02d}'.format(dt.month),
        'M': lambda dt: '{:d}'.format(dt.month),

        # Day
        'DD': lambda dt: '{:02d}'.format(dt.day),
        'D': lambda dt: '{:d}'.format(dt.day),

        # Day of Year
        'DDDD': lambda dt: '{:03d}'.format(dt.day_of_year),
        'DDD': lambda dt: '{:d}'.format(dt.day_of_year),

        # Day of Week
        'd': lambda dt: '{:d}'.format(dt.day_of_week),

        # Hour
        'HH': lambda dt: '{:02d}'.format(dt.hour),
        'H': lambda dt: '{:d}'.format(dt.hour),
        'hh': lambda dt: '{:02d}'.format(dt.hour % 12 or 12),
        'h': lambda dt: '{:d}'.format(dt.hour % 12 or 12),

        # Minute
        'mm': lambda dt: '{:02d}'.format(dt.minute),
        'm': lambda dt: '{:d}'.format(dt.minute),

        # Second
        'ss': lambda dt: '{:02d}'.format(dt.second),
        's': lambda dt: '{:d}'.format(dt.second),

        # Fractional second
        'S': lambda dt: '{:01d}'.format(dt.microsecond // 100000),
        'SS': lambda dt: '{:02d}'.format(dt.microsecond // 10000),
        'SSS': lambda dt: '{:03d}'.format(dt.microsecond // 1000),
        'SSSS': lambda dt: '{:04d}'.format(dt.microsecond // 100),
        'SSSSS': lambda dt: '{:05d}'.format(dt.microsecond // 10),
        'SSSSSS': lambda dt: '{:06d}'.format(dt.microsecond),

        # Timestamp
        'X': lambda dt: '{:d}'.format(dt.int_timestamp),

        # Timezone
        'z': lambda dt: '{}'.format(dt.tzinfo.abbrev),
        'zz': lambda dt: '{}'.format(dt.timezone_name),
    }

    _DEFAULT_DATE_FORMATS = {
        'LTS': 'h:mm:ss A',
        'LT': 'h:mm A',
        'L': 'MM/DD/YYYY',
        'LL': 'MMMM D, YYYY',
        'LLL': 'MMMM D, YYYY h:mm A',
        'LLLL': 'dddd, MMMM D, YYYY h:mm A',
    }

    _REGEX_TOKENS = {
        'Y': _MATCH_SIGNED,
        'YY': (_MATCH_1_TO_2, _MATCH_2),
        'YYYY': (_MATCH_1_TO_4, _MATCH_4),
        'Q': _MATCH_1,
        'Qo': None,
        'M': _MATCH_1_TO_2,
        'MM': (_MATCH_1_TO_2, _MATCH_2),
        'MMM': None,
        'MMMM': None,
        'D': _MATCH_1_TO_2,
        'DD': (_MATCH_1_TO_2, _MATCH_2),
        'DDD': _MATCH_1_TO_3,
        'DDDD': _MATCH_3,
        'Do': None,
        'H': _MATCH_1_TO_2,
        'HH': (_MATCH_1_TO_2, _MATCH_2),
        'h': _MATCH_1_TO_2,
        'hh': (_MATCH_1_TO_2, _MATCH_2),
        'm': _MATCH_1_TO_2,
        'mm': (_MATCH_1_TO_2, _MATCH_2),
        's': _MATCH_1_TO_2,
        'ss': (_MATCH_1_TO_2, _MATCH_2),
        'S': (_MATCH_1_TO_3, _MATCH_1),
        'SS': (_MATCH_1_TO_3, _MATCH_2),
        'SSS': (_MATCH_1_TO_3, _MATCH_3),
        'SSSS': _MATCH_UNSIGNED,
        'SSSSS': _MATCH_UNSIGNED,
        'SSSSSS': _MATCH_UNSIGNED,
        'a': None,
        'x': _MATCH_SIGNED,
        'X': re.compile('[+-]?\d+(\.\d{1,3})?')
    }

    _PARSE_TOKENS = {
        'YYYY': lambda year: int(year),
        'YY': lambda year: 1900 + int(year),
        'Q': lambda quarter: int(quarter),
        'MMMM': lambda month: None,
        'MMM': lambda month: None,
        'MM': lambda month: int(month),
        'M': lambda month: int(month),
        'DDDD': lambda day: int(day),
        'DDD': lambda day: int(day),
        'DD': lambda day: int(day),
        'D': lambda day: int(day),
        'HH': lambda hour: int(hour),
        'H': lambda hour: int(hour),
        'hh': lambda hour: int(hour),
        'h': lambda hour: int(hour),
        'mm': lambda minute: int(minute),
        'm': lambda minute: int(minute),
        'ss': lambda second: int(second),
        's': lambda second: int(second),
        'S': lambda us: int(us) * 100000,
        'SS': lambda us: int(us) * 10000,
        'SSS': lambda us: int(us) * 1000,
        'SSSS': lambda us: int(us) * 100,
        'SSSSS': lambda us: int(us) * 10,
        'SSSSSS': lambda us: int(us),
        'a': lambda meridiem: None,
        'X': lambda ts: float(ts),
        'x': lambda ts: float(ts) / 1e3,
    }

    def format(self, dt, fmt, locale=None):
        """
        Formats a Pendulum instance with a given format and locale.

        :param dt: The instance to format
        :type dt: pendulum.Pendulum

        :param fmt: The format to use
        :type fmt: str

        :param locale: The locale to use
        :type locale: str or None

        :rtype: str
        """
        if not locale:
            locale = dt.get_locale()

        return self._FORMAT_RE.sub(
            lambda m: m.group(1)
                if m.group(1)
                else m.group(2)
                if m.group(2)
                else self._format_token(dt, m.group(3), locale),
            fmt
        )

    def _format_token(self, dt, token, locale):
        """
        Formats a Pendulum instance with a given token and locale.

        :param dt: The instance to format
        :type dt: pendulum.Pendulum

        :param token: The token to use
        :type token: str

        :param locale: The locale to use
        :type locale: str or None

        :rtype: str
        """
        if token in self._DEFAULT_DATE_FORMATS:
            fmt = dt.translator().transchoice('date_formats', token, locale=locale)
            if fmt == 'date_formats':
                fmt = self._DEFAULT_DATE_FORMATS[token]

            return self.format(dt, fmt, locale)
        if token in self._LOCALIZABLE_TOKENS:
            return self._format_localizable_token(dt, token, locale)

        if token in self._TOKENS_RULES:
            return self._TOKENS_RULES[token](dt)

        # Timezone
        if token in ['ZZ', 'Z']:
            separator = ':' if token == 'ZZ' else ''
            offset = dt.utcoffset() or datetime.timedelta()
            minutes = offset.total_seconds() / 60

            if minutes >= 0:
                sign = '+'
            else:
                sign = '-'

            hour, minute = divmod(abs(int(minutes)), 60)

            return '{}{:02d}{}{:02d}'.format(sign, hour, separator, minute)

    def _format_localizable_token(self, dt, token, locale):
        """
        Formats a Pendulum instance
        with a given localizable token and locale.

        :param dt: The instance to format
        :type dt: pendulum.Pendulum

        :param token: The token to use
        :type token: str

        :param locale: The locale to use
        :type locale: str or None

        :rtype: str
        """
        trans_id = ''
        count = 0

        if token == 'MMM':
            count = dt.month
            trans_id = 'months_abbrev'
        elif token == 'MMMM':
            count = dt.month
            trans_id = 'months'
        elif token in ('dd', 'ddd'):
            count = dt.day_of_week
            trans_id = 'days_abbrev'
        elif token == 'dddd':
            count = dt.day_of_week
            trans_id = 'days'
        elif token == 'Do':
            count = dt.day
            trans_id = 'ordinal'
        elif token == 'do':
            count = dt.day_of_week
            trans_id = 'ordinal'
        elif token == 'Mo':
            count = dt.month
            trans_id = 'ordinal'
        elif token == 'Qo':
            count = dt.quarter
            trans_id = 'ordinal'
        elif token == 'wo':
            count = dt.week_of_year
            trans_id = 'ordinal'
        elif token == 'DDDo':
            count = dt.day_of_year
            trans_id = 'ordinal'
        elif token == 'A':
            count = (dt.hour, dt.minute)
            trans_id = 'meridian'

        trans = dt.translator().transchoice(trans_id, count, locale=locale)

        if trans_id == 'ordinal':
            trans = '{:d}{}'.format(count, trans)

        if trans_id == trans:
            # Unable to find the corresponding translation
            # Defaulting to english
            return self._format_localizable_token(dt, token, 'en')

        return decode(trans)

    def parse(self, time, fmt):
        """
        Parses a time string matching a given format as a tuple.

        :param time: The timestring
        :type time: str

        :param fmt: The format
        :type fmt: str

        :rtype: tuple
        """
        fmt = re.escape(fmt)
        tokens = self._FORMAT_RE.findall(fmt)
        if not tokens:
            return time

        parsed = {
            'year': None,
            'month': None,
            'day': None,
            'hour': None,
            'minute': None,
            'second': None,
            'microsecond': None,
            'tz': None,
            'quarter': None,
            'day_of_week': None,
            'day_of_year': None,
            'meridiem': None,
            'timestamp': None
        }

        pattern = self._FORMAT_RE.sub(lambda m: self._replace_tokens(m.group(0)), fmt)

        if not re.match(pattern, time):
            raise ValueError('String does not match format {}'.format(fmt))

        re.sub(pattern, lambda m: self._get_parsed_values(m, parsed), time)

        return parsed

    def _get_parsed_values(self, m, parsed):
        for token, index in m.re.groupindex.items():
            self._get_parsed_value(token, m.group(index), parsed)

    def _get_parsed_value(self, token, value, parsed):
        parsed_token = self._PARSE_TOKENS[token](value)

        if 'Y' in token:
            parsed['year'] = parsed_token
        elif 'Q' == token:
            parsed['quarter'] = parsed_token
        elif 'M' in token:
            parsed['month'] = parsed_token
        elif token in ['DDDD', 'DDD']:
            parsed['day_of_year'] = parsed_token
        elif 'D' in token:
            parsed['day'] = parsed_token
        elif 'H' in token:
            parsed['hour'] = parsed_token
        elif token in ['hh', 'h']:
            parsed['hour'] = parsed_token
        elif 'm' in token:
            parsed['minute'] = parsed_token
        elif 's' in token:
            parsed['second'] = parsed_token
        elif 'S' in token:
            parsed['microsecond'] = parsed_token
        elif token in ['MMM', 'MMMM']:
            parsed['day_of_week'] = parsed_token
        elif token == 'a':
            pass
        elif token in ['X', 'x']:
            parsed['timestamp'] = parsed_token

    def _replace_tokens(self, token):
        if token.startswith('[') and token.endswith(']'):
            return token[1:-1]
        elif token.startswith('\\'):
            return token
        elif token not in self._REGEX_TOKENS:
            raise ValueError('Unsupported token: {}'.format(token))

        candidates = self._REGEX_TOKENS[token]
        if not candidates:
            raise ValueError('Unsupported token: {}'.format(token))

        if not isinstance(candidates, tuple):
            candidates = (candidates,)

        pattern = '(?P<{}>{})'.format(token, '|'.join([p.pattern for p in candidates]))

        return pattern



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'Sondag',
        1: 'Maandag',
        2: 'Dinsdag',
        3: 'Woensdag',
        4: 'Donderdag',
        5: 'Vrydag',
        6: 'Saterdag'
    },
    'days_abbrev': {
        0: 'Son',
        1: 'Maa',
        2: 'Din',
        3: 'Woe',
        4: 'Don',
        5: 'Vry',
        6: 'Sat'
    },

    # Months
    'months': {
        1: 'Januarie',
        2: 'Februarie',
        3: 'Maart',
        4: 'April',
        5: 'Mei',
        6: 'Junie',
        7: 'Julie',
        8: 'Augustus',
        9: 'September',
        10: 'Oktober',
        11: 'November',
        12: 'Desember',
    },
    'months_abbrev': {
        1: 'Jan',
        2: 'Feb',
        3: 'Mrt',
        4: 'Apr',
        5: 'Mei',
        6: 'Jun',
        7: 'Jul',
        8: 'Aug',
        9: 'Sep',
        10: 'Okt',
        11: 'Nov',
        12: 'Des',
    },

    # Units of time
    'year': ['{count} jaar', '{count} jare'],
    'month': ['{count} maand', '{count} maande'],
    'week': ['{count} week', '{count} weke'],
    'day': ['{count} dag', '{count} dae'],
    'hour': ['{count} uur', '{count} ure'],
    'minute': ['{count} minuut', '{count} minute'],
    'second': ['{count} sekond', '{count} sekondes'],

    # Relative time
    'ago': '{time} terug',
    'from_now': '{time} van nou af',
    'after': '{time} na',
    'before': '{time} voor',

    # Meridians
    'meridian': lambda time: 'VM' if 0 <= time[0] < 12 else 'NM',

    # Date formats
    'date_formats': {
        'LTS': 'HH:mm:ss',
        'LT': 'HH:mm',
        'LLLL': 'dddd, D MMMM YYYY HH:mm',
        'LLL': 'D MMMM YYYY HH:mm',
        'LL': 'D MMMM YYYY',
        'L': 'DD/MM/YYYY',
    },
}




# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'Chủ Nhật',
        1: 'Thứ Hai',
        2: 'Thứ Ba',
        3: 'Thứ Tư',
        4: 'Thứ Năm',
        5: 'Thứ Sáu',
        6: 'Thứ Bảy'
    },
    'days_abbrev': {
        0: 'CN',
        1: 'Thứ 2',
        2: 'Thứ 3',
        3: 'Thứ 4',
        4: 'Thứ 5',
        5: 'Thứ 6',
        6: 'Thứ 7'
    },

    # Months
    'months': {
        1: 'Tháng Một',
        2: 'Tháng Hai',
        3: 'Tháng Ba',
        4: 'Tháng Tư',
        5: 'Tháng Năm',
        6: 'Tháng Sáu',
        7: 'Tháng Bảy',
        8: 'Tháng Tám',
        9: 'Tháng Chín',
        10: 'Tháng Mười',
        11: 'Tháng Mười Một',
        12: 'Tháng Mười Hai',
    },
    'months_abbrev': {
        1: 'Tháng 1',
        2: 'Tháng 2',
        3: 'Tháng 3',
        4: 'Tháng 4',
        5: 'Tháng 5',
        6: 'Tháng 6',
        7: 'Tháng 7',
        8: 'Tháng 8',
        9: 'Tháng 9',
        10: 'Tháng 10',
        11: 'Tháng 11',
        12: 'Tháng 12',
    },

    # Units of time
    'year': '{count} năm',
    'month': '{count} tháng',
    'week': '{count} tuần',
    'day': '{count} ngày',
    'hour': '{count} giờ',
    'minute': '{count} phút',
    'second': '{count} giây',

    # Relative time
    'ago': '{time} trước',
    'from_now': '{time} từ bây giờ',
    'after': '{time} sau',
    'before': '{time} trước',

    # Date formats
    'date_formats': {
        'LTS': 'HH:mm:ss',
        'LT': 'HH:mm',
        'LLLL': 'dddd, D MMMM [năm] YYYY HH:mm',
        'LLL': 'D MMMM [năm] YYYY HH:mm',
        'LL': 'D MMMM [năm] YYYY',
        'L': 'DD/MM/YYYY',
    },
}



# -*- coding: utf-8 -*-
import math

from ..constants import (
    EPOCH_YEAR,
    SECS_PER_DAY,
    SECS_PER_400_YEARS,
    SECS_PER_100_YEARS,
    SECS_PER_4_YEARS,
    SECS_PER_YEAR,
    SECS_PER_HOUR,
    SECS_PER_MIN,
    MONTHS_OFFSETS,
    TM_DECEMBER,
    TM_JANUARY
)


def local_time(unix_time, utc_offset, microseconds):
    """
    Returns a UNIX time as a broken down time
    for a particular transition type.

    :type unix_time: int
    :type utc_offset: int
    :type microseconds: int

    :rtype: tuple
    """
    year = EPOCH_YEAR
    seconds = int(math.floor(unix_time))

    # Shift to a base year that is 400-year aligned.
    if seconds >= 0:
        seconds -= 10957 * SECS_PER_DAY
        year += 30  # == 2000
    else:
        seconds += (146097 - 10957) * SECS_PER_DAY
        year -= 370  # == 1600

    seconds += utc_offset

    # Handle years in chunks of 400/100/4/1
    year += 400 * (seconds // SECS_PER_400_YEARS)
    seconds %= SECS_PER_400_YEARS
    if seconds < 0:
        seconds += SECS_PER_400_YEARS
        year -= 400

    leap_year = 1  # 4-century aligned

    sec_per_100years = SECS_PER_100_YEARS[leap_year]
    while seconds >= sec_per_100years:
        seconds -= sec_per_100years
        year += 100
        leap_year = 0  # 1-century, non 4-century aligned
        sec_per_100years = SECS_PER_100_YEARS[leap_year]

    sec_per_4years = SECS_PER_4_YEARS[leap_year]
    while seconds >= sec_per_4years:
        seconds -= sec_per_4years
        year += 4
        leap_year = 1  # 4-year, non century aligned
        sec_per_4years = SECS_PER_4_YEARS[leap_year]

    sec_per_year = SECS_PER_YEAR[leap_year]
    while seconds >= sec_per_year:
        seconds -= sec_per_year
        year += 1
        leap_year = 0  # non 4-year aligned
        sec_per_year = SECS_PER_YEAR[leap_year]

    # Handle months and days
    month = TM_DECEMBER + 1
    day = seconds // SECS_PER_DAY + 1
    seconds %= SECS_PER_DAY
    while month != TM_JANUARY + 1:
        month_offset = MONTHS_OFFSETS[leap_year][month]
        if day > month_offset:
            day -= month_offset
            break

        month -= 1

    # Handle hours, minutes, seconds and microseconds
    hour = seconds // SECS_PER_HOUR
    seconds %= SECS_PER_HOUR
    minute = seconds // SECS_PER_MIN
    second = seconds % SECS_PER_MIN

    return (
        year, month, day,
        hour, minute, second, microseconds
    )



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'sunnuntai',
        1: 'maanantai',
        2: 'tiistai',
        3: 'keskiviikko',
        4: 'torstai',
        5: 'perjantai',
        6: 'lauantai'
    },
    'days_abbrev': {
        0: 'su',
        1: 'ma',
        2: 'ti',
        3: 'ke',
        4: 'to',
        5: 'pe',
        6: 'la'
    },

    # Months
    'months': {
        1: 'tammikuu',
        2: 'helmikuu',
        3: 'maaliskuu',
        4: 'huhtikuu',
        5: 'toukokuu',
        6: 'kesäkuu',
        7: 'heinäkuu',
        8: 'elokuu',
        9: 'syyskuu',
        10: 'lokakuu',
        11: 'marraskuu',
        12: 'joulukuu',
    },
    'months_abbrev': {
        1: 'tammi',
        2: 'helmi',
        3: 'maalis',
        4: 'huhti',
        5: 'touko',
        6: 'kesä',
        7: 'heinä',
        8: 'elo',
        9: 'syys',
        10: 'loka',
        11: 'marras',
        12: 'joulu',
    },

    # Units of time
    'year': ['{count} vuosi', '{count} vuotta'],
    'month': ['{count} kuukausi', '{count} kuukautta'],
    'week': ['{count} viikko', '{count} viikkoa'],
    'day': ['{count} päivä', '{count} päivää'],
    'hour': ['{count} tunti', '{count} tuntia'],
    'minute': ['{count} minuutti', '{count} minuuttia'],
    'second': ['{count} sekunti', '{count} sekuntia'],

    # Relative time
    'ago': '{time} sitten',
    'from_now': '{time} tästä hetkestä',
    'after': '{time} sen jälkeen',
    'before': '{time} ennen',

    # Date formats
    'date_formats': {
        'LTS': 'HH.mm.ss',
        'LT': 'HH.mm',
        'LLLL': 'dddd, Do MMMM[ta] YYYY, [klo] HH.mm',
        'LLL': 'Do MMMM[ta] YYYY, [klo] HH.mm',
        'LL': 'Do MMMM[ta] YYYY',
        'L': 'DD.MM.YYYY',
    },
}



# -*- coding: utf-8 -*-

from __future__ import division

import calendar
import datetime
import warnings

import pendulum

from .date import Date
from .time import Time
from .period import Period
from .exceptions import PendulumException
from .tz import Timezone, UTC, FixedTimezone, local_timezone
from .tz.timezone_info import TimezoneInfo
from .parsing import parse
from .helpers import add_duration
from .formatting import FORMATTERS
from .constants import (
    YEARS_PER_CENTURY, YEARS_PER_DECADE,
    MONTHS_PER_YEAR,
    MINUTES_PER_HOUR, SECONDS_PER_MINUTE,
    SECONDS_PER_DAY,
    SUNDAY, SATURDAY
)


class Pendulum(Date, datetime.datetime):

    # Formats
    ATOM = '%Y-%m-%dT%H:%M:%S%_z'
    COOKIE = '%A, %d-%b-%Y %H:%M:%S %Z'
    ISO8601 = '%Y-%m-%dT%H:%M:%S%_z'
    ISO8601_EXTENDED = '%Y-%m-%dT%H:%M:%S.%f%_z'
    RFC822 = '%a, %d %b %y %H:%M:%S %z'
    RFC850 = '%A, %d-%b-%y %H:%M:%S %Z'
    RFC1036 = '%a, %d %b %y %H:%M:%S %z'
    RFC1123 = '%a, %d %b %Y %H:%M:%S %z'
    RFC2822 = '%a, %d %b %Y %H:%M:%S %z'
    RFC3339 = '%Y-%m-%dT%H:%M:%S%_z'
    RFC3339_EXTENDED = '%Y-%m-%dT%H:%M:%S.%f%_z'
    RSS = '%a, %d %b %Y %H:%M:%S %z'
    W3C = '%Y-%m-%dT%H:%M:%S%_z'

    _EPOCH = datetime.datetime(1970, 1, 1, tzinfo=UTC)

    _TRANSITION_RULE = Timezone.POST_TRANSITION

    _MODIFIERS_VALID_UNITS = [
        'second', 'minute', 'hour',
        'day', 'week', 'month', 'year',
        'decade', 'century'
    ]

    @classmethod
    def _safe_create_datetime_zone(cls, obj):
        """
        Creates a timezone from a string, Timezone, TimezoneInfo or integer offset.

        :param obj: str or Timezone or TimezoneInfo or int or None

        :rtype: Timezone
        """
        if isinstance(obj, TimezoneInfo):
            return obj.tz

        if obj is None or obj == 'local':
            return cls._local_timezone()

        if isinstance(obj, (int, float)):
            timezone_offset = obj * 60 * 60

            return FixedTimezone.load(timezone_offset)
        elif isinstance(obj, datetime.tzinfo) and not isinstance(obj, Timezone):
            # pytz
            if hasattr(obj, 'localize'):
                return cls._timezone(obj.zone)

            return FixedTimezone.load(obj.utcoffset(None).total_seconds())

        return cls._timezone(obj)

    @classmethod
    def _local_timezone(cls):
        """
        Returns the local timezone using tzlocal.get_localzone
        or the TZ environment variable.

        :rtype: Timezone
        """
        return local_timezone()

    @classmethod
    def _timezone(cls, zone):
        """
        Returns a timezone given its name.

        :param zone: The name of the timezone.
        :type zone: str

        :rtype: Timezone
        """
        if isinstance(zone, Timezone):
            return zone

        return Timezone.load(zone)

    def __new__(cls, year, month, day,
                hour=0, minute=0, second=0, microsecond=0,
                tzinfo=None, fold=None):
        """
        Constructor.

        This will just create a dummy datetime. The heavy lifting will be done
        in __init__().

        :type year: int or str
        """
        obj = super(Pendulum, cls).__new__(cls, year, month, day, hour, minute, second, microsecond, None)

        return obj

    def __init__(self, year, month, day,
                 hour=0, minute=0, second=0, microsecond=0,
                 tzinfo=UTC, fold=None):
        # If a TimezoneInfo is passed we do not convert
        if isinstance(tzinfo, TimezoneInfo):
            self._tz = tzinfo.tz

            self._year = year
            self._month = month
            self._day = day
            self._hour = hour
            self._minute = minute
            self._second = second
            self._microsecond = microsecond
            self._tzinfo = tzinfo

            if fold is None:
                # Converting rule to fold value
                if self._TRANSITION_RULE == Timezone.POST_TRANSITION:
                    fold = 1
                else:
                    fold = 0

            self._fold = fold

            dt = datetime.datetime(
                year, month, day,
                hour, minute, second, microsecond,
                tzinfo
            )
        else:
            self._tz = self._safe_create_datetime_zone(tzinfo)

            # Support for explicit fold attribute
            if fold is None:
                transition_rule = self._TRANSITION_RULE

                # Converting rule to fold value
                if self._TRANSITION_RULE == Timezone.POST_TRANSITION:
                    fold = 1
                else:
                    fold = 0
            elif fold == 1:
                transition_rule = Timezone.POST_TRANSITION
            else:
                transition_rule = Timezone.PRE_TRANSITION

            dt = self._tz.convert(datetime.datetime(
                year, month, day,
                hour, minute, second, microsecond
            ), dst_rule=transition_rule)

            self._year = dt.year
            self._month = dt.month
            self._day = dt.day
            self._hour = dt.hour
            self._minute = dt.minute
            self._second = dt.second
            self._microsecond = dt.microsecond
            self._tzinfo = dt.tzinfo
            self._fold = fold

        self._timestamp = None
        self._int_timestamp = None
        self._datetime = dt

    @classmethod
    def instance(cls, dt, tz=UTC):
        """
        Create a Pendulum instance from a datetime one.

        :param dt: A datetime instance
        :type dt: datetime.datetime

        :param tz: The timezone
        :type tz: Timezone or TimeZoneInfo or str or None

        :rtype: Pendulum
        """
        tz = dt.tzinfo or tz

        # Checking for pytz/tzinfo
        if isinstance(tz, datetime.tzinfo) and not isinstance(tz, (Timezone, TimezoneInfo)):
            # pytz
            if hasattr(tz, 'localize') and tz.zone:
                tz = tz.zone
            else:
                # We have no sure way to figure out
                # the timezone name, we fallback
                # on a fixed offset
                tz = tz.utcoffset(dt).total_seconds() / 3600

        return cls(
            dt.year, dt.month, dt.day,
            dt.hour, dt.minute, dt.second, dt.microsecond,
            tzinfo=tz
        )

    @classmethod
    def parse(cls, time=None, tz=UTC, **options):
        """
        Create a Pendulum instance from a string.

        :param time: The time string
        :type time: str

        :param tz: The timezone
        :type tz: Timezone or TimezoneInfo or str or None

        :rtype: Pendulum
        """
        if time is None:
            return cls.now(tz)

        if time == 'now':
            return cls.now(None)

        parsed = parse(time, **options)

        if parsed['offset'] is None:
            tz = tz
        else:
            tz = parsed['offset'] / 3600

        return cls(
            parsed['year'], parsed['month'], parsed['day'],
            parsed['hour'], parsed['minute'], parsed['second'],
            parsed['subsecond'],
            tzinfo=tz
        )

    @classmethod
    def now(cls, tz=None):
        """
        Get a Pendulum instance for the current date and time.

        :param tz: The timezone
        :type tz: Timezone or TimezoneInfo or str or int or None

        :rtype: Pendulum
        """
        # If the class has a test now set and we are trying to create a now()
        # instance then override as required
        if cls.has_test_now():
            test_instance = cls.get_test_now()

            if tz is not None and tz != test_instance.timezone:
                test_instance = test_instance.in_timezone(tz)

            return test_instance

        if tz is None or tz == 'local':
            dt = datetime.datetime.now(cls._local_timezone())
        elif tz is UTC or tz == 'UTC':
            dt = datetime.datetime.now(UTC)
        else:
            dt = datetime.datetime.now(UTC)
            tz = cls._safe_create_datetime_zone(tz)
            dt = tz.convert(dt)

        return cls.instance(dt, tz)

    @classmethod
    def utcnow(cls):
        """
        Get a Pendulum instance for the current date and time in UTC.

        :param tz: The timezone
        :type tz: Timezone or TimezoneInfo or str or None

        :rtype: Pendulum
        """
        return cls.now(UTC)

    @classmethod
    def today(cls, tz=None):
        """
        Create a Pendulum instance for today.

        :param tz: The timezone
        :type tz: Timezone or TimezoneInfo or str or None

        :rtype: Pendulum
        """
        return cls.now(tz).start_of('day')

    @classmethod
    def tomorrow(cls, tz=None):
        """
        Create a Pendulum instance for tomorrow.

        :param tz: The timezone
        :type tz: Timezone or TimezoneInfo or str or None

        :rtype: Pendulum
        """
        return cls.today(tz).add(days=1)

    @classmethod
    def yesterday(cls, tz=None):
        """
        Create a Pendulum instance for yesterday.

        :param tz: The timezone
        :type tz: Timezone or TimezoneInfo or str or None

        :rtype: Pendulum
        """
        return cls.today(tz).subtract(days=1)

    @classmethod
    def create(cls, year=None, month=None, day=None,
               hour=0, minute=0, second=0, microsecond=0,
               tz=UTC):
        """
        Create a new Pendulum instance from a specific date and time.

        If any of year, month or day are set to None their now() values will
        be used.

        :type year: int
        :type month: int
        :type day: int
        :type hour: int
        :type minute: int
        :type second: int
        :type microsecond: int
        :type tz: tzinfo or str or int or None

        :rtype: Pendulum
        """
        tz = cls._safe_create_datetime_zone(tz)

        if any([year is None, month is None, day is None]):
            if cls.has_test_now():
                now = cls.get_test_now().in_tz(tz)
            else:
                now = datetime.datetime.now(UTC)
                now = tz.convert(now)

            if year is None:
                year = now.year

            if month is None:
                month = now.month

            if day is None:
                day = now.day

        dt = datetime.datetime(
            year, month, day, hour, minute, second, microsecond
        )

        return cls.instance(dt, tz)

    @classmethod
    def create_from_format(cls, time, fmt, tz=UTC, formatter='classic'):
        """
        Create a Pendulum instance from a specific format.

        :param fmt: The format
        :type fmt: str

        :param time: The time string
        :type time: str

        :param tz: The timezone
        :type tz: tzinfo or str or int or None

        :param formatter: The formatter to use. Default "classic"
        :type formatter: str

        :rtype: Pendulum
        """
        if formatter not in FORMATTERS:
            raise ValueError('Invalid formatter [{}]'.format(formatter))

        if formatter == 'classic':
            warnings.warn(
                'Using the classic formatter in from_format() '
                'is deprecated and will no longer be possible '
                'in version 2.0. Use the alternative formatter instead.',
                DeprecationWarning,
                stacklevel=2
            )
            dt = datetime.datetime.strptime(time, fmt)

            return cls.instance(dt, tz)

        formatter = FORMATTERS['alternative']

        parts = formatter.parse(time, fmt)
        actual_parts = {}

        # If timestamp has been specified
        # we use it and don't go any further
        if parts['timestamp'] is not None:
            return cls.create_from_timestamp(parts['timestamp'])

        if parts['quarter'] is not None:
            dt = cls.now().start_of('year')

            while dt.quarter != parts['quarter']:
                dt = dt.add(months=3)

            actual_parts['year'] = dt.year
            actual_parts['month'] = dt.month
            actual_parts['day'] = dt.day

        # If the date part has not been specified
        # we default to today
        if all([parts['year'] is None, parts['month'] is None, parts['day'] is None]):
            now = cls.now()

            parts['year'] = actual_parts['year'] = now.year
            parts['month'] = actual_parts['month'] = now.month
            parts['day'] = actual_parts['day'] = now.day

        # We replace any not set month/day value
        # with the first of each unit
        if any([parts['month'] is None, parts['day'] is None]):
            for part in ['month', 'day']:
                if parts[part] is None and actual_parts.get(part) is None:
                    actual_parts[part] = 1

        for part in ['year', 'month', 'day']:
            if parts[part] is not None:
                actual_parts[part] = parts[part]

        # If any of hour/minute/second/microsecond is not set
        # We assume start of corresponding value
        for part in ['hour', 'minute', 'second', 'microsecond']:
            if parts[part] is None:
                actual_parts[part] = 0
            else:
                actual_parts[part] = parts[part]

        if parts['day_of_year'] is not None:
            text = '{}-{}'.format(
                actual_parts['year'], parts['day_of_year']
            )
            dt = parse(text)

            actual_parts['month'] = dt['month']
            actual_parts['day'] = dt['day']

        # Meridiem
        if parts['meridiem'] is not None:
            pass

        actual_parts['tz'] = parts['tz'] or tz

        return cls.create(**actual_parts)

    @classmethod
    def create_from_timestamp(cls, timestamp, tz=UTC):
        """
        Create a Pendulum instance from a timestamp.

        :param timestamp: The timestamp
        :type timestamp: int or float

        :param tz: The timezone
        :type tz: Timezone or TimezoneInfo or str or int or None

        :rtype: Pendulum
        """
        dt = datetime.datetime.utcfromtimestamp(timestamp).replace(tzinfo=UTC)
        instance = cls(
            dt.year, dt.month, dt.day,
            dt.hour, dt.minute, dt.second, dt.microsecond,
            dt.tzinfo
        )

        if tz is not UTC and tz != 'UTC':
            instance = instance.in_tz(tz)

        return instance

    @classmethod
    def strptime(cls, time, fmt):
        dt = datetime.datetime.strptime(time, fmt)

        return cls.instance(dt)

    def copy(self):
        """
        Get a copy of the instance.

        :rtype: Pendulum
        """
        return self.instance(self._datetime)

    # Getters/Setters

    def hour_(self, hour):
        return self._setter(hour=hour)

    def minute_(self, minute):
        return self._setter(minute=minute)

    def second_(self, second):
        return self._setter(second=second)

    def microsecond_(self, microsecond):
        return self._setter(microsecond=microsecond)

    def _setter(self, **kwargs):
        kwargs['tzinfo'] = True

        return self._tz.convert(self.replace(**kwargs))

    def timezone_(self, tz):
        return self.__class__(
            self.year, self.month, self.day,
            self.hour, self.minute, self.second, self.microsecond,
            tz
        )

    def tz_(self, tz):
        return self.timezone_(tz)

    def timestamp_(self, timestamp, tz=UTC):
        return self.create_from_timestamp(timestamp, tz)

    @property
    def year(self):
        return self._year

    @property
    def month(self):
        return self._month

    @property
    def day(self):
        return self._day

    @property
    def hour(self):
        return self._hour

    @property
    def minute(self):
        return self._minute

    @property
    def second(self):
        return self._second

    @property
    def microsecond(self):
        return self._microsecond

    @property
    def tzinfo(self):
        return self._tzinfo

    @property
    def fold(self):
        return self._fold

    def timestamp(self):
        if self._timestamp is None:
            delta = self._datetime - self._EPOCH

            self._timestamp = delta.total_seconds()

        return self._timestamp

    @property
    def float_timestamp(self):
        return self.timestamp()

    @property
    def int_timestamp(self):
        if self._int_timestamp is None:
            delta = self._datetime - self._EPOCH

            self._int_timestamp = delta.days * SECONDS_PER_DAY + delta.seconds

        return self._int_timestamp

    @property
    def offset(self):
        return self.get_offset()

    @property
    def offset_hours(self):
        return (self.get_offset()
                / SECONDS_PER_MINUTE
                / MINUTES_PER_HOUR)

    @property
    def local(self):
        return self.offset == self.in_timezone(self._local_timezone()).offset

    @property
    def utc(self):
        return self.offset == 0

    @property
    def is_dst(self):
        return self.tzinfo.is_dst

    @property
    def timezone(self):
        return self.get_timezone()

    @property
    def tz(self):
        return self.get_timezone()

    @property
    def timezone_name(self):
        return self.timezone.name

    @property
    def age(self):
        return self.date().diff(self.now(self._tz).date()).in_years()

    def get_timezone(self):
        return self._tz

    def get_offset(self):
        return int(self._tzinfo.offset)

    def date(self):
        return Date.instance(self._datetime.date())

    def time(self):
        return Time(self.hour, self.minute, self.second, self.microsecond)

    def on(self, year, month, day):
        """
        Returns a new instance with the current date set to a different date.

        :param year: The year
        :type year: int

        :param month: The month
        :type month: int

        :param day: The day
        :type day: int

        :rtype: Pendulum
        """
        return self.replace(
            year=int(year), month=int(month), day=int(day)
        )

    def at(self, hour, minute, second, microsecond=0):
        """
        Returns a new instance with the current time to a different time.

        :param hour: The hour
        :type hour: int

        :param minute: The minute
        :type minute: int

        :param second: The second
        :type second: int

        :param microsecond: The microsecond
        :type microsecond: int

        :rtype: Pendulum
        """
        return self.replace(
            hour=hour, minute=minute, second=second,
            microsecond=microsecond
        )

    def with_date_time(self, year, month, day, hour, minute, second, microsecond=0):
        """
        Return a new instance with the date and time set to the given values.

        :type year: int
        :type month: int
        :type day: int
        :type hour: int
        :type minute: int
        :type second: int

        :rtype: Pendulum
        """
        return self.replace(
            year=year, month=month, day=day,
            hour=hour, minute=minute, second=second,
            microsecond=microsecond
        )

    def with_time_from_string(self, time):
        """
        Returns a new instance with the time set by time string.

        :param time: The time string
        :type time: str

        :rtype: Pendulum
        """
        time = time.split(':')

        hour = int(time[0])
        minute = int(time[1]) if len(time) > 1 else 0
        second = int(time[2]) if len(time) > 2 else 0

        return self.at(hour, minute, second)

    def in_timezone(self, tz):
        """
        Set the instance's timezone from a string or object.

        :param value: The timezone
        :type value: Timezone or TimezoneInfo or str or None

        :rtype: Pendulum
        """
        tz = self._safe_create_datetime_zone(tz)

        return tz.convert(self)

    def in_tz(self, tz):
        """
        Set the instance's timezone from a string or object.

        :param value: The timezone
        :type value: Timezone or TimezoneInfo or str or int or None

        :rtype: Pendulum
        """
        return self.in_timezone(tz)

    def with_timestamp(self, timestamp):
        """
        Set the date and time based on a Unix timestamp.

        :param timestamp: The timestamp
        :type timestamp: int or float
        :rtype: Pendulum
        """
        dt = datetime.datetime.fromtimestamp(timestamp, UTC).astimezone(self._tz)

        return self.instance(dt)

    # Normalization Rule
    @classmethod
    def set_transition_rule(cls, rule):
        if rule not in [Timezone.PRE_TRANSITION,
                        Timezone.POST_TRANSITION,
                        Timezone.TRANSITION_ERROR]:
            raise ValueError('Invalid transition rule: {}'.format(rule))

        cls._TRANSITION_RULE = rule

    @classmethod
    def get_transition_rule(cls):
        return cls._TRANSITION_RULE

    # STRING FORMATTING

    def to_time_string(self):
        """
        Format the instance as time.

        :rtype: str
        """
        return self.format('%H:%M:%S', formatter='classic')

    def to_datetime_string(self):
        """
        Format the instance as date and time.

        :rtype: str
        """
        return self.format('%Y-%m-%d %H:%M:%S', formatter='classic')

    def to_day_datetime_string(self):
        """
        Format the instance as day, date and time.

        :rtype: str
        """
        return self.format('ddd, MMM D, YYYY h:mm A', formatter='alternative')

    def to_atom_string(self):
        """
        Format the instance as ATOM.

        :rtype: str
        """
        return self.format(self.ATOM, formatter='classic')

    def to_cookie_string(self):
        """
        Format the instance as COOKIE.

        :rtype: str
        """
        return self.format(self.COOKIE, formatter='classic')

    def to_iso8601_string(self, extended=False):
        """
        Format the instance as ISO8601.

        :rtype: str
        """
        fmt = self.ISO8601
        if extended:
            fmt = self.ISO8601_EXTENDED

        return self.format(fmt, formatter='classic')

    def to_rfc822_string(self):
        """
        Format the instance as RFC822.

        :rtype: str
        """
        return self.format(self.RFC822, formatter='classic')

    def to_rfc850_string(self):
        """
        Format the instance as RFC850.

        :rtype: str
        """
        return self.format(self.RFC850, formatter='classic')

    def to_rfc1036_string(self):
        """
        Format the instance as RFC1036.

        :rtype: str
        """
        return self.format(self.RFC1036, formatter='classic')

    def to_rfc1123_string(self):
        """
        Format the instance as RFC1123.

        :rtype: str
        """
        return self.format(self.RFC1123, formatter='classic')

    def to_rfc2822_string(self):
        """
        Format the instance as RFC2822.

        :rtype: str
        """
        return self.format(self.RFC2822, formatter='classic')

    def to_rfc3339_string(self, extended=False):
        """
        Format the instance as RFC3339.

        :rtype: str
        """
        fmt = self.RFC3339
        if extended:
            fmt = self.RFC3339_EXTENDED

        return self.format(fmt, formatter='classic')

    def to_rss_string(self):
        """
        Format the instance as RSS.

        :rtype: str
        """
        return self.format(self.RSS, formatter='classic')

    def to_w3c_string(self):
        """
        Format the instance as W3C.

        :rtype: str
        """
        return self.format(self.W3C, formatter='classic')

    # Comparisons
    def __eq__(self, other):
        try:
            return self._datetime == self._get_datetime(other)
        except ValueError:
            return NotImplemented

    def __ne__(self, other):
        try:
            return self._datetime != self._get_datetime(other)
        except ValueError:
            return NotImplemented

    def __gt__(self, other):
        try:
            return self._datetime > self._get_datetime(other)
        except ValueError:
            return NotImplemented

    def __ge__(self, other):
        try:
            return self._datetime >= self._get_datetime(other)
        except ValueError:
            return NotImplemented

    def __lt__(self, other):
        try:
            return self._datetime < self._get_datetime(other)
        except ValueError:
            return NotImplemented

    def __le__(self, other):
        try:
            return self._datetime <= self._get_datetime(other)
        except ValueError:
            return NotImplemented

    def between(self, dt1, dt2, equal=True):
        """
        Determines if the instance is between two others.

        :type dt1: Pendulum or datetime or str or int
        :type dt2: Pendulum or datetime or str or int

        :param equal: Indicates if a > and < comparison shoud be used or <= and >=

        :rtype: bool
        """
        if dt1 > dt2:
            dt1, dt2 = dt2, dt1

        if equal:
            return self >= dt1 and self <= dt2

        return self > dt1 and self < dt2

    def closest(self, dt1, dt2):
        """
        Get the closest date from the instance.

        :type dt1: Pendulum or datetime
        :type dt2: Pendulum or datetime

        :rtype: Pendulum
        """
        dt1 = self._get_datetime(dt1, True)
        dt2 = self._get_datetime(dt2, True)

        if self.diff(dt1).in_seconds() < self.diff(dt2).in_seconds():
            return dt1

        return dt2

    def farthest(self, dt1, dt2):
        """
        Get the farthest date from the instance.

        :type dt1: Pendulum or datetime
        :type dt2: Pendulum or datetime

        :rtype: Pendulum
        """
        dt1 = self._get_datetime(dt1, True)
        dt2 = self._get_datetime(dt2, True)

        if self.diff(dt1).in_seconds() > self.diff(dt2).in_seconds():
            return dt1

        return dt2

    def min_(self, dt=None):
        """
        Get the minimum instance between a given instance (default utcnow)
        and the current instance.

        :type dt: Pendulum or datetime or str or int

        :rtype: Pendulum
        """
        if dt is None:
            dt = Pendulum.now(self.timezone)

        if self < dt:
            return self

        return self._get_datetime(dt, True)

    def minimum(self, dt=None):
        """
        Get the minimum instance between a given instance (default now)
        and the current instance.

        :type dt: Pendulum or datetime or str or int

        :rtype: Pendulum
        """
        return self.min_(dt)

    def max_(self, dt=None):
        """
        Get the maximum instance between a given instance (default now)
        and the current instance.

        :type dt: Pendulum or datetime or str or int

        :rtype: Pendulum
        """
        if dt is None:
            dt = Pendulum.now(self.timezone)

        if self > dt:
            return self

        return self._get_datetime(dt, True)

    def maximum(self, dt=None):
        """
        Get the maximum instance between a given instance (default utcnow)
        and the current instance.

        :type dt: Pendulum or datetime or str or int

        :rtype: Pendulum
        """
        return self.max_(dt)

    def is_yesterday(self):
        """
        Determines if the instance is yesterday.

        :rtype: bool
        """
        return self.to_date_string() == self.yesterday(self._tz).to_date_string()

    def is_today(self):
        """
        Determines if the instance is today.

        :rtype: bool
        """
        return self.to_date_string() == self.now(self._tz).to_date_string()

    def is_tomorrow(self):
        """
        Determines if the instance is tomorrow.

        :rtype: bool
        """
        return self.to_date_string() == self.tomorrow(self._tz).to_date_string()

    def is_future(self):
        """
        Determines if the instance is in the future, ie. greater than now.

        :rtype: bool
        """
        return self > self.now(self.timezone)

    def is_past(self):
        """
        Determines if the instance is in the past, ie. less than now.

        :rtype: bool
        """
        return self < self.now(self.timezone)

    def is_long_year(self):
        """
        Determines if the instance is a long year

        See link `https://en.wikipedia.org/wiki/ISO_8601#Week_dates`_

        :rtype: bool
        """
        return Pendulum.create(self.year, 12, 28, 0, 0, 0, tz=self._tz).isocalendar()[1] == 53

    def is_same_day(self, dt):
        """
        Checks if the passed in date is the same day as the instance current day.

        :type dt: Pendulum or datetime or str or int

        :rtype: bool
        """
        dt = self._get_datetime(dt, True)

        return self.to_date_string() == dt.to_date_string()

    def is_birthday(self, dt=None):
        """
        Check if its the birthday. Compares the date/month values of the two dates.

        :rtype: bool
        """
        if dt is None:
            dt = Pendulum.now(self._tz)

        instance = self._get_datetime(dt, True)

        return (self.month, self.day) == (instance.month, instance.day)

    # ADDITIONS AND SUBSTRACTIONS

    def add(self, years=0, months=0, weeks=0, days=0,
            hours=0, minutes=0, seconds=0, microseconds=0):
        """
        Add duration to the instance.

        :param years: The number of years
        :type years: int

        :param months: The number of months
        :type months: int

        :param weeks: The number of weeks
        :type weeks: int

        :param days: The number of days
        :type days: int

        :param hours: The number of hours
        :type hours: int

        :param minutes: The number of minutes
        :type minutes: int

        :param seconds: The number of seconds
        :type seconds: int

        :param microseconds: The number of microseconds
        :type microseconds: int

        :rtype: Pendulum
        """
        dt = add_duration(
            self._datetime,
            years=years, months=months, weeks=weeks, days=days,
            hours=hours, minutes=minutes, seconds=seconds,
            microseconds=microseconds
        )

        if any([years, months, weeks, days]):
            # If we specified any of years, months, weeks or days
            # we will not apply the transition (if any)
            return self.__class__(
                dt.year, dt.month, dt.day,
                dt.hour, dt.minute, dt.second, dt.microsecond,
                tzinfo=self._tz,
                fold=1
            )

        # Else, we need to apply the transition properly (if any)
        dt = self._tz.convert(dt)

        return self.instance(dt)

    def subtract(self, years=0, months=0, weeks=0, days=0,
                 hours=0, minutes=0, seconds=0, microseconds=0):
        """
        Remove duration from the instance.

        :param years: The number of years
        :type years: int

        :param months: The number of months
        :type months: int

        :param weeks: The number of weeks
        :type weeks: int

        :param days: The number of days
        :type days: int

        :param hours: The number of hours
        :type hours: int

        :param minutes: The number of minutes
        :type minutes: int

        :param seconds: The number of seconds
        :type seconds: int

        :param microseconds: The number of microseconds
        :type microseconds: int

        :rtype: Pendulum
        """
        return self.add(
            years=-years, months=-months, weeks=-weeks, days=-days,
            hours=-hours, minutes=-minutes, seconds=-seconds,
            microseconds=-microseconds
        )

    def add_timedelta(self, delta):
        """
        Add timedelta duration to the instance.

        :param delta: The timedelta instance
        :type delta: datetime.timedelta

        :rtype: Pendulum
        """
        if isinstance(delta, Period):
            return self.add(
                years=delta.years, months=delta.months,
                weeks=delta.weeks, days=delta.remaining_days,
                hours=delta.hours, minutes=delta.minutes,
                seconds=delta.remaining_seconds, microseconds=delta.microseconds
            )

        return self.add(seconds=delta.total_seconds())

    def subtract_timedelta(self, delta):
        """
        Remove timedelta duration from the instance.

        :param delta: The timedelta instance
        :type delta: datetime.timedelta

        :rtype: Pendulum
        """
        if isinstance(delta, Period):
            return self.subtract(
                years=delta.years, months=delta.months,
                weeks=delta.weeks, days=delta.remaining_days,
                hours=delta.hours, minutes=delta.minutes,
                seconds=delta.remaining_seconds, microseconds=delta.microseconds
            )

        return self.subtract(days=delta.days, seconds=delta.seconds,
                             microseconds=delta.microseconds)

    # DIFFERENCES

    def seconds_since_midnight(self):
        """
        The number of seconds since midnight.

        :rtype: int
        """
        return self.diff(self.start_of('day')).in_seconds()

    def seconds_until_end_of_day(self):
        """
        The number of seconds until 23:59:59.

        :rtype: int
        """
        return self.diff(self.end_of('day')).in_seconds()

    def diff(self, dt=None, abs=True):
        """
        Returns the difference between two Pendulum objects represented as a Interval.

        :type dt: Pendulum or None

        :param abs: Whether to return an absolute interval or not
        :type abs: bool

        :rtype: Period
        """
        if dt is None:
            dt = self.now(self._tz)

        return Period(self, self._get_datetime(dt, pendulum=True), absolute=abs)

    # Modifiers
    def start_of(self, unit):
        """
        Returns a copy of the instance with the time reset
        with the following rules:

        * second: microsecond set to 0
        * minute: second and microsecond set to 0
        * hour: minute, second and microsecond set to 0
        * day: time to 00:00:00
        * week: date to first day of the week and time to 00:00:00
        * month: date to first day of the month and time to 00:00:00
        * year: date to first day of the year and time to 00:00:00
        * decade: date to first day of the decade and time to 00:00:00
        * century: date to first day of century and time to 00:00:00

        :param unit: The unit to reset to
        :type unit: str

        :rtype: Pendulum
        """
        if unit not in self._MODIFIERS_VALID_UNITS:
            raise ValueError('Invalid unit "{}" for start_of()'.format(unit))

        return getattr(self, '_start_of_{}'.format(unit))()

    def end_of(self, unit):
        """
        Returns a copy of the instance with the time reset
        with the following rules:

        * second: microsecond set to 999999
        * minute: second set to 59 and microsecond set to 999999
        * hour: minute and second set to 59 and microsecond set to 999999
        * day: time to 23:59:59.999999
        * week: date to last day of the week and time to 23:59:59.999999
        * month: date to last day of the month and time to 23:59:59.999999
        * year: date to last day of the year and time to 23:59:59.999999
        * decade: date to last day of the decade and time to 23:59:59.999999
        * century: date to last day of century and time to 23:59:59.999999

        :param unit: The unit to reset to
        :type unit: str

        :rtype: Pendulum
        """
        if unit not in self._MODIFIERS_VALID_UNITS:
            raise ValueError('Invalid unit "%s" for end_of()' % unit)

        return getattr(self, '_end_of_%s' % unit)()

    def _start_of_second(self):
        """
        Reset microseconds to 0.

        :rtype: Pendulum
        """
        return self.microsecond_(0)

    def _end_of_second(self):
        """
        Set microseconds to 999999.

        :rtype: Pendulum
        """
        return self.microsecond_(999999)

    def _start_of_minute(self):
        """
        Reset seconds and microseconds to 0.

        :rtype: Pendulum
        """
        return self.replace(second=0, microsecond=0)

    def _end_of_minute(self):
        """
        Set seconds to 59 and microseconds to 999999.

        :rtype: Pendulum
        """
        return self.replace(second=59, microsecond=999999)

    def _start_of_hour(self):
        """
        Reset minutes, seconds and microseconds to 0.

        :rtype: Pendulum
        """
        return self.replace(minute=0, second=0, microsecond=0)

    def _end_of_hour(self):
        """
        Set minutes and seconds to 59 and microseconds to 999999.

        :rtype: Pendulum
        """
        return self.replace(minute=59, second=59, microsecond=999999)

    def _start_of_day(self):
        """
        Reset the time to 00:00:00

        :rtype: Pendulum
        """
        return self.at(0, 0, 0)

    def _end_of_day(self):
        """
        Reset the time to 23:59:59.999999

        :rtype: Pendulum
        """
        return self.at(23, 59, 59, 999999)

    def _start_of_month(self):
        """
        Reset the date to the first day of the month and the time to 00:00:00.

        :rtype: Pendulum
        """
        return self.with_date_time(self.year, self.month, 1, 0, 0, 0)

    def _end_of_month(self):
        """
        Reset the date to the last day of the month
        and the time to 23:59:59.999999.

        :rtype: Pendulum
        """
        return self.with_date_time(
            self.year, self.month, self.days_in_month, 23, 59, 59, 999999
        )

    def _start_of_year(self):
        """
        Reset the date to the first day of the year and the time to 00:00:00.

        :rtype: Pendulum
        """
        return self.with_date_time(self.year, 1, 1, 0, 0, 0)

    def _end_of_year(self):
        """
        Reset the date to the last day of the year
        and the time to 23:59:59.999999

        :rtype: Pendulum
        """
        return self.with_date_time(
            self.year, 12, 31, 23, 59, 59, 999999
        )

    def _start_of_decade(self):
        """
        Reset the date to the first day of the decade
        and the time to 00:00:00.

        :rtype: Pendulum
        """
        year = self.year - self.year % YEARS_PER_DECADE
        return self.with_date_time(year, 1, 1, 0, 0, 0)

    def _end_of_decade(self):
        """
        Reset the date to the last day of the decade
        and the time to 23:59:59.999999.

        :rtype: Pendulum
        """
        year = self.year - self.year % YEARS_PER_DECADE + YEARS_PER_DECADE - 1

        return self.with_date_time(
            year, 12, 31, 23, 59, 59, 999999
        )

    def _start_of_century(self):
        """
        Reset the date to the first day of the century
        and the time to 00:00:00.

        :rtype: Pendulum
        """
        year = self.year - 1 - (self.year - 1) % YEARS_PER_CENTURY + 1

        return self.with_date_time(year, 1, 1, 0, 0, 0)

    def _end_of_century(self):
        """
        Reset the date to the last day of the century
        and the time to 23:59:59.999999.

        :rtype: Pendulum
        """
        year = self.year - 1 - (self.year - 1) % YEARS_PER_CENTURY + YEARS_PER_CENTURY

        return self.with_date_time(
            year, 12, 31, 23, 59, 59, 999999
        )

    def _start_of_week(self):
        """
        Reset the date to the first day of the week
        and the time to 00:00:00.

        :rtype: Pendulum
        """
        dt = self

        if self.day_of_week != self._week_starts_at:
            dt = self.previous(self._week_starts_at)

        return dt.start_of('day')

    def _end_of_week(self):
        """
        Reset the date to the last day of the week
        and the time to 23:59:59.

        :rtype: Pendulum
        """
        dt = self

        if self.day_of_week != self._week_ends_at:
            dt = self.next(self._week_ends_at)

        return dt.end_of('day')

    def next(self, day_of_week=None, keep_time=False):
        """
        Modify to the next occurrence of a given day of the week.
        If no day_of_week is provided, modify to the next occurrence
        of the current day of the week.  Use the supplied consts
        to indicate the desired day_of_week, ex. Pendulum.MONDAY.

        :param day_of_week: The next day of week to reset to.
        :type day_of_week: int or None

        :param keep_time: Whether to keep the time information or not.
        :type keep_time: bool

        :rtype: Pendulum
        """
        if day_of_week is None:
            day_of_week = self.day_of_week

        if day_of_week < SUNDAY or day_of_week > SATURDAY:
            raise ValueError('Invalid day of week')

        if keep_time:
            dt = self
        else:
            dt = self.start_of('day')

        dt = dt.add(days=1)
        while dt.day_of_week != day_of_week:
            dt = dt.add(days=1)

        return dt

    def previous(self, day_of_week=None, keep_time=False):
        """
        Modify to the previous occurrence of a given day of the week.
        If no day_of_week is provided, modify to the previous occurrence
        of the current day of the week.  Use the supplied consts
        to indicate the desired day_of_week, ex. Pendulum.MONDAY.

        :param day_of_week: The previous day of week to reset to.
        :type day_of_week: int or None

        :param keep_time: Whether to keep the time information or not.
        :type keep_time: bool

        :rtype: Pendulum
        """
        if day_of_week is None:
            day_of_week = self.day_of_week

        if day_of_week < SUNDAY or day_of_week > SATURDAY:
            raise ValueError('Invalid day of week')

        if keep_time:
            dt = self
        else:
            dt = self.start_of('day')

        dt = dt.subtract(days=1)
        while dt.day_of_week != day_of_week:
            dt = dt.subtract(days=1)

        return dt

    def first_of(self, unit, day_of_week=None):
        """
        Returns an instance set to the first occurrence
        of a given day of the week in the current unit.
        If no day_of_week is provided, modify to the first day of the unit.
        Use the supplied consts to indicate the desired day_of_week, ex. Pendulum.MONDAY.

        Supported units are month, quarter and year.

        :param unit: The unit to use
        :type unit: str

        :type day_of_week: int or None

        :rtype: Pendulum
        """
        if unit not in ['month', 'quarter', 'year']:
            raise ValueError('Invalid unit "{}" for first_of()'.format(unit))

        return getattr(self, '_first_of_{}'.format(unit))(day_of_week)

    def last_of(self, unit, day_of_week=None):
        """
        Returns an instance set to the last occurrence
        of a given day of the week in the current unit.
        If no day_of_week is provided, modify to the last day of the unit.
        Use the supplied consts to indicate the desired day_of_week, ex. Pendulum.MONDAY.

        Supported units are month, quarter and year.

        :param unit: The unit to use
        :type unit: str

        :type day_of_week: int or None

        :rtype: Pendulum
        """
        if unit not in ['month', 'quarter', 'year']:
            raise ValueError('Invalid unit "{}" for first_of()'.format(unit))

        return getattr(self, '_last_of_{}'.format(unit))(day_of_week)

    def nth_of(self, unit, nth, day_of_week):
        """
        Returns a new instance set to the given occurrence
        of a given day of the week in the current unit.
        If the calculated occurrence is outside the scope of the current unit,
        then raise an error. Use the supplied consts
        to indicate the desired day_of_week, ex. Pendulum.MONDAY.

        Supported units are month, quarter and year.

        :param unit: The unit to use
        :type unit: str

        :type nth: int

        :type day_of_week: int or None

        :rtype: Pendulum
        """
        if unit not in ['month', 'quarter', 'year']:
            raise ValueError('Invalid unit "{}" for first_of()'.format(unit))

        dt = getattr(self, '_nth_of_{}'.format(unit))(nth, day_of_week)
        if dt is False:
            raise PendulumException('Unable to find occurence {} of {} in {}'.format(
                                     nth, self._days[day_of_week], unit))

        return dt

    def _first_of_month(self, day_of_week):
        """
        Modify to the first occurrence of a given day of the week
        in the current month. If no day_of_week is provided,
        modify to the first day of the month. Use the supplied consts
        to indicate the desired day_of_week, ex. Pendulum.MONDAY.

        :type day_of_week: int

        :rtype: Pendulum
        """
        dt = self.start_of('day')

        if day_of_week is None:
            return dt.day_(1)

        month = calendar.monthcalendar(dt.year, dt.month)

        calendar_day = (day_of_week - 1) % 7

        if month[0][calendar_day] > 0:
            day_of_month = month[0][calendar_day]
        else:
            day_of_month = month[1][calendar_day]

        return dt.day_(day_of_month)

    def _last_of_month(self, day_of_week=None):
        """
        Modify to the last occurrence of a given day of the week
        in the current month. If no day_of_week is provided,
        modify to the last day of the month. Use the supplied consts
        to indicate the desired day_of_week, ex. Pendulum.MONDAY.

        :type day_of_week: int or None

        :rtype: Pendulum
        """
        dt = self.start_of('day')

        if day_of_week is None:
            return dt.day_(self.days_in_month)

        month = calendar.monthcalendar(dt.year, dt.month)

        calendar_day = (day_of_week - 1) % 7

        if month[-1][calendar_day] > 0:
            day_of_month = month[-1][calendar_day]
        else:
            day_of_month = month[-2][calendar_day]

        return dt.day_(day_of_month)

    def _nth_of_month(self, nth, day_of_week):
        """
        Modify to the given occurrence of a given day of the week
        in the current month. If the calculated occurrence is outside,
        the scope of the current month, then return False and no
        modifications are made. Use the supplied consts
        to indicate the desired day_of_week, ex. Pendulum.MONDAY.

        :type nth: int

        :type day_of_week: int or None

        :rtype: Pendulum
        """
        if nth == 1:
            return self.first_of('month', day_of_week)

        dt = self.first_of('month')
        check = dt.format('%Y-%m')
        for i in range(nth - (1 if dt.day_of_week == day_of_week else 0)):
            dt = dt.next(day_of_week)

        if dt.format('%Y-%m') == check:
            return self.day_(dt.day).start_of('day')

        return False

    def _first_of_quarter(self, day_of_week=None):
        """
        Modify to the first occurrence of a given day of the week
        in the current quarter. If no day_of_week is provided,
        modify to the first day of the quarter. Use the supplied consts
        to indicate the desired day_of_week, ex. Pendulum.MONDAY.

        :type day_of_week: int or None

        :rtype: Pendulum
        """
        return self.on(self.year, self.quarter * 3 - 2, 1).first_of('month', day_of_week)

    def _last_of_quarter(self, day_of_week=None):
        """
        Modify to the last occurrence of a given day of the week
        in the current quarter. If no day_of_week is provided,
        modify to the last day of the quarter. Use the supplied consts
        to indicate the desired day_of_week, ex. Pendulum.MONDAY.

        :type day_of_week: int or None

        :rtype: Pendulum
        """
        return self.on(self.year, self.quarter * 3, 1).last_of('month', day_of_week)

    def _nth_of_quarter(self, nth, day_of_week):
        """
        Modify to the given occurrence of a given day of the week
        in the current quarter. If the calculated occurrence is outside,
        the scope of the current quarter, then return False and no
        modifications are made. Use the supplied consts
        to indicate the desired day_of_week, ex. Pendulum.MONDAY.

        :type nth: int

        :type day_of_week: int or None

        :rtype: Pendulum
        """
        if nth == 1:
            return self.first_of('quarter', day_of_week)

        dt = self.day_(1).month_(self.quarter * 3)
        last_month = dt.month
        year = dt.year
        dt = dt.first_of('quarter')
        for i in range(nth - (1 if dt.day_of_week == day_of_week else 0)):
            dt = dt.next(day_of_week)

        if last_month < dt.month or year != dt.year:
            return False

        return self.on(self.year, dt.month, dt.day).start_of('day')

    def _first_of_year(self, day_of_week=None):
        """
        Modify to the first occurrence of a given day of the week
        in the current year. If no day_of_week is provided,
        modify to the first day of the year. Use the supplied consts
        to indicate the desired day_of_week, ex. Pendulum.MONDAY.

        :type day_of_week: int or None

        :rtype: Pendulum
        """
        return self.month_(1).first_of('month', day_of_week)

    def _last_of_year(self, day_of_week=None):
        """
        Modify to the last occurrence of a given day of the week
        in the current year. If no day_of_week is provided,
        modify to the last day of the year. Use the supplied consts
        to indicate the desired day_of_week, ex. Pendulum.MONDAY.

        :type day_of_week: int or None

        :rtype: Pendulum
        """
        return self.month_(MONTHS_PER_YEAR).last_of('month', day_of_week)

    def _nth_of_year(self, nth, day_of_week):
        """
        Modify to the given occurrence of a given day of the week
        in the current year. If the calculated occurrence is outside,
        the scope of the current year, then return False and no
        modifications are made. Use the supplied consts
        to indicate the desired day_of_week, ex. Pendulum.MONDAY.

        :type nth: int

        :type day_of_week: int or None

        :rtype: Pendulum
        """
        if nth == 1:
            return self.first_of('year', day_of_week)

        dt = self.first_of('year')
        year = dt.year
        for i in range(nth - (1 if dt.day_of_week == day_of_week else 0)):
            dt = dt.next(day_of_week)

        if year != dt.year:
            return False

        return self.on(self.year, dt.month, dt.day).start_of('day')

    def average(self, dt=None):
        """
        Modify the current instance to the average
        of a given instance (default now) and the current instance.

        :type dt: Pendulum or datetime

        :rtype: Pendulum
        """
        if dt is None:
            dt = Pendulum.now(self._tz)

        diff = self.diff(dt, False)
        return self.add(microseconds=(diff.in_seconds() * 1000000 + diff.microseconds) // 2)

    def _get_datetime(self, value, pendulum=False):
        """
        Gets a datetime from a given value.

        :param value: The value to get the datetime from.
        :type value: Pendulum or datetime or int or float or str.

        :param pendulum: Whether to return a Pendulum instance.
        :type pendulum: bool

        :rtype: datetime or Pendulum
        """
        if value is None:
            return None

        if isinstance(value, Pendulum):
            return value._datetime if not pendulum else value

        if isinstance(value, datetime.datetime):
            if value.tzinfo is None:
                value = self._tz.convert(value)

            return value if not pendulum else Pendulum.instance(value)

        raise ValueError('Invalid datetime "{}"'.format(value))

    def __sub__(self, other):
        if isinstance(other, datetime.timedelta):
            return self.subtract_timedelta(other)

        try:
            return self._get_datetime(other, True).diff(self, False)
        except ValueError:
            return NotImplemented

    def __rsub__(self, other):
        try:
            return self.diff(self._get_datetime(other, True), False)
        except ValueError:
            return NotImplemented

    def __add__(self, other):
        if not isinstance(other, datetime.timedelta):
            return NotImplemented

        return self.add_timedelta(other)

    def __radd__(self, other):
        return self.__add__(other)

    # Native methods override

    @classmethod
    def fromtimestamp(cls, t, tz=None):
        return cls.create_from_timestamp(t, tz)

    @classmethod
    def utcfromtimestamp(cls, t):
        return cls.create_from_timestamp(t)

    @classmethod
    def fromordinal(cls, n):
        return cls.instance(datetime.datetime.fromordinal(n))

    @classmethod
    def combine(cls, date, time):
        return cls.instance(datetime.datetime.combine(date, time))

    def timetuple(self):
        return self._datetime.timetuple()

    def utctimetuple(self):
        return self._datetime.utctimetuple()

    def replace(self, year=None, month=None, day=None, hour=None,
                minute=None, second=None, microsecond=None, tzinfo=True,
                fold=None):
        year = year if year is not None else self._year
        month = month if month is not None else self._month
        day = day if day is not None else self._day
        hour = hour if hour is not None else self._hour
        minute = minute if minute is not None else self._minute
        second = second if second is not None else self._second
        microsecond = microsecond if microsecond is not None else self._microsecond

        # Checking tzinfo
        if tzinfo is not None and tzinfo is not True:
            tzinfo = self._safe_create_datetime_zone(tzinfo)
        elif tzinfo is None:
            tzinfo = UTC
        else:
            tzinfo = self._tzinfo.tz

        return self.__class__(
            year, month, day,
            hour, minute, second, microsecond,
            tzinfo=tzinfo, fold=fold
        )

    def astimezone(self, tz=None):
        tz = self._safe_create_datetime_zone(tz)

        return self.instance(self._datetime.astimezone(tz))

    def isoformat(self, sep='T'):
        return self._datetime.isoformat(sep)

    def utcoffset(self):
        return self._tzinfo.utcoffset(self)

    def tzname(self):
        return self._datetime.tzname()

    def dst(self):
        return self._datetime.dst()

    def __hash__(self):
        return self._datetime.__hash__()

    def __getnewargs__(self):
        return(self, )

    def _getstate(self, protocol=3):
        return (
            self.year, self.month, self.day,
            self.hour, self.minute, self.second, self.microsecond,
            self.tzinfo,
            self.fold
        )

    def __reduce__(self):
        return self.__reduce_ex__(2)

    def __reduce_ex__(self, protocol):
        return self.__class__, self._getstate(protocol)

Pendulum.min = Pendulum.instance(datetime.datetime.min.replace(tzinfo=UTC))
Pendulum.max = Pendulum.instance(datetime.datetime.max.replace(tzinfo=UTC))
Pendulum.EPOCH = Pendulum(1970, 1, 1)



# -*- coding: utf-8 -*-

# The day constants
SUNDAY = 0
MONDAY = 1
TUESDAY = 2
WEDNESDAY = 3
THURSDAY = 4
FRIDAY = 5
SATURDAY = 6

# Number of X in Y.
YEARS_PER_CENTURY = 100
YEARS_PER_DECADE = 10
MONTHS_PER_YEAR = 12
WEEKS_PER_YEAR = 52
DAYS_PER_WEEK = 7
HOURS_PER_DAY = 24
MINUTES_PER_HOUR = 60
SECONDS_PER_MINUTE = 60
SECONDS_PER_HOUR = MINUTES_PER_HOUR * SECONDS_PER_MINUTE
SECONDS_PER_DAY = HOURS_PER_DAY * SECONDS_PER_HOUR

# Formats
ATOM = '%Y-%m-%dT%H:%M:%S%_z'
COOKIE = '%A, %d-%b-%Y %H:%M:%S %Z'
ISO8601 = '%Y-%m-%dT%H:%M:%S%_z'
ISO8601_EXTENDED = '%Y-%m-%dT%H:%M:%S.%f%_z'
RFC822 = '%a, %d %b %y %H:%M:%S %z'
RFC850 = '%A, %d-%b-%y %H:%M:%S %Z'
RFC1036 = '%a, %d %b %y %H:%M:%S %z'
RFC1123 = '%a, %d %b %Y %H:%M:%S %z'
RFC2822 = '%a, %d %b %Y %H:%M:%S %z'
RFC3339 = '%Y-%m-%dT%H:%M:%S%_z'
RFC3339_EXTENDED = '%Y-%m-%dT%H:%M:%S.%f%_z'
RSS = '%a, %d %b %Y %H:%M:%S %z'
W3C = '%Y-%m-%dT%H:%M:%S%_z'


EPOCH_YEAR = 1970

DAYS_PER_N_YEAR = 365
DAYS_PER_L_YEAR = 366

USECS_PER_SEC = 1000000

SECS_PER_MIN = 60
SECS_PER_HOUR = 60 * SECS_PER_MIN
SECS_PER_DAY = SECS_PER_HOUR * 24

# 400-year chunks always have 146097 days (20871 weeks).
SECS_PER_400_YEARS = 146097 * SECS_PER_DAY

# The number of seconds in an aligned 100-year chunk, for those that
# do not begin with a leap year and those that do respectively.
SECS_PER_100_YEARS = (
    (76 * DAYS_PER_N_YEAR + 24 * DAYS_PER_L_YEAR) * SECS_PER_DAY,
    (75 * DAYS_PER_N_YEAR + 25 * DAYS_PER_L_YEAR) * SECS_PER_DAY
)

# The number of seconds in an aligned 4-year chunk, for those that
# do not begin with a leap year and those that do respectively.
SECS_PER_4_YEARS = (
    (4 * DAYS_PER_N_YEAR + 0 * DAYS_PER_L_YEAR) * SECS_PER_DAY,
    (3 * DAYS_PER_N_YEAR + 1 * DAYS_PER_L_YEAR) * SECS_PER_DAY
)

# The number of seconds in non-leap and leap years respectively.
SECS_PER_YEAR = (
    DAYS_PER_N_YEAR * SECS_PER_DAY,
    DAYS_PER_L_YEAR * SECS_PER_DAY
)

# The month lengths in non-leap and leap years respectively.
DAYS_PER_MONTHS = (
    (-1, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31),
    (-1, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)
)

# The day offsets of the beginning of each (1-based) month in non-leap
# and leap years respectively.
# For example, in a leap year there are 335 days before December.
MONTHS_OFFSETS = (
    (-1, 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365),
    (-1, 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366)
)

DAY_OF_WEEK_TABLE = (
    0, 3, 2, 5, 0, 3, 5, 1, 4, 6, 2, 4
)

TM_SUNDAY = 0
TM_MONDAY = 1
TM_TUESDAY = 2
TM_WEDNESDAY = 3
TM_THURSDAY = 4
TM_FRIDAY = 5
TM_SATURDAY = 6

TM_JANUARY = 0
TM_FEBRUARY = 1
TM_MARCH = 2
TM_APRIL = 3
TM_MAY = 4
TM_JUNE = 5
TM_JULY = 6
TM_AUGUST = 7
TM_SEPTEMBER = 8
TM_OCTOBER = 9
TM_NOVEMBER = 10
TM_DECEMBER = 11



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'воскресенье',
        1: 'понедельник',
        2: 'вторник',
        3: 'среда',
        4: 'четверг',
        5: 'пятница',
        6: 'суббота'
    },
    'days_abbrev': {
        0: 'вс',
        1: 'пн',
        2: 'вт',
        3: 'ср',
        4: 'чт',
        5: 'пт',
        6: 'сб'
    },

    # Months
    'months': {
        1: 'января',
        2: 'февраля',
        3: 'марта',
        4: 'апреля',
        5: 'мая',
        6: 'июня',
        7: 'июля',
        8: 'августа',
        9: 'сентября',
        10: 'октября',
        11: 'ноября',
        12: 'декабря',
    },
    'months_abbrev': {
        1: 'янв',
        2: 'фев',
        3: 'мар',
        4: 'апр',
        5: 'май',
        6: 'июн',
        7: 'июл',
        8: 'авг',
        9: 'сен',
        10: 'окт',
        11: 'ноя',
        12: 'дек',
    },

    # Units of time
    'year': ['{count} год', '{count} года', '{count} лет'],
    'month': ['{count} месяц', '{count} месяца', '{count} месяцев'],
    'week': ['{count} неделю', '{count} недели', '{count} недель'],
    'day': ['{count} день', '{count} дня', '{count} дней'],
    'hour': ['{count} час', '{count} часа', '{count} часов'],
    'minute': ['{count} минуту', '{count} минуты', '{count} минут'],
    'second': ['{count} секунду', '{count} секунды', '{count} секунд'],

    # Relative time
    'ago': '{time} назад',
    'from_now': 'через {time}',
    'after': '{time} после',
    'before': '{time} до',

    # Date formats
    'date_formats': {
        'LTS': 'HH:mm:ss',
        'LT': 'HH:mm',
        'LLLL': 'dddd, D MMMM YYYY г., HH:mm',
        'LLL': 'D MMMM YYYY г., HH:mm',
        'LL': 'D MMMM YYYY г.',
        'L': 'DD.MM.YYYY',
    },
}



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'dimanche',
        1: 'lundi',
        2: 'mardi',
        3: 'mercredi',
        4: 'jeudi',
        5: 'vendredi',
        6: 'samedi'
    },
    'days_abbrev': {
        0: 'dim',
        1: 'lun',
        2: 'mar',
        3: 'mer',
        4: 'jeu',
        5: 'ven',
        6: 'sam'
    },

    # Months
    'months': {
        1: 'janvier',
        2: 'février',
        3: 'mars',
        4: 'avril',
        5: 'mai',
        6: 'juin',
        7: 'juillet',
        8: 'août',
        9: 'septembre',
        10: 'octobre',
        11: 'novembre',
        12: 'décembre',
    },
    'months_abbrev': {
        1: 'janv',
        2: 'févr',
        3: 'mars',
        4: 'avr',
        5: 'mai',
        6: 'juin',
        7: 'juil',
        8: 'août',
        9: 'sept',
        10: 'oct',
        11: 'nov',
        12: 'déc',
    },

    # Units of time
    'year': ['{count} an', '{count} ans'],
    'month': '{count} mois',
    'week': ['{count} semaine', '{count} semaines'],
    'day': ['{count} jour', '{count} jours'],
    'hour': ['{count} heure', '{count} heures'],
    'minute': ['{count} minute', '{count} minutes'],
    'second': ['{count} seconde', '{count} secondes'],

    # Relative Time
    'ago': 'il y a {time}',
    'from_now': 'dans {time}',
    'after': '{time} après',
    'before': '{time} avant',

    # Ordinals
    'ordinal': {
        1: 'er',
        'default': 'e'
    },

    # Date formats
    'date_formats': {
        'LTS': 'HH:mm:ss',
        'LT': 'HH:mm',
        'LLLL': 'dddd D MMMM YYYY HH:mm',
        'LLL': 'D MMMM YYYY HH:mm',
        'LL': 'D MMMM YYYY',
        'L': 'DD/MM/YYYY',
    },
}



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'pühapäev',
        1: 'esmaspäev',
        2: 'teisipäev',
        3: 'kolmapäev',
        4: 'neljapäev',
        5: 'reede',
        6: 'laupäev'
    },
    'days_abbrev': {
        0: 'P',
        1: 'E',
        2: 'T',
        3: 'K',
        4: 'N',
        5: 'R',
        6: 'L'
    },

    # Months
    'months': {
        1: 'jaanuar',
        2: 'veebruar',
        3: 'märts',
        4: 'aprill',
        5: 'mai',
        6: 'juuni',
        7: 'juuli',
        8: 'august',
        9: 'september',
        10: 'oktoober',
        11: 'november',
        12: 'detsember',
    },
    'months_abbrev': {
        1: 'jaan',
        2: 'veebr',
        3: 'märts',
        4: 'apr',
        5: 'mai',
        6: 'juuni',
        7: 'juuli',
        8: 'aug',
        9: 'sept',
        10: 'okt',
        11: 'nov',
        12: 'dets',
    },

    # Units of time
    'year': ['{count} aasta', '{count} aastat'],
    'month': ['{count} kuu', '{count} kuud'],
    'week': ['{count} nädal', '{count} nädalat'],
    'day': ['{count} päev', '{count} päeva'],
    'hour': ['{count} tund', '{count} tundi'],
    'minute': ['{count} minut', '{count} minutit'],
    'second': ['{count} sekund', '{count} sekundit'],

    # Relative time
    'ago': '{time} tagasi',
    'from_now': '{time} pärast',
    'after': '{time} pärast',
    'before': '{time} enne',
    'year_from_now': '{count} aasta',
    'month_from_now': '{count} kuu',
    'week_from_now': '{count} nädala',
    'day_from_now': '{count} päeva',
    'hour_from_now': '{count} tunni',
    'minute_from_now': '{count} minuti',
    'second_from_now': '{count} sekundi',

    # Date formats
    'date_formats': {
        'LTS': 'H:mm:ss',
        'LLLL': 'dddd, D. MMMM YYYY H:mm',
        'LLL': 'dddd, D. MMMM YYYY H:mm',
        'LL': 'dddd, D. MMMM YYYY H:mm',
        'L': 'dddd, D. MMMM YYYY H:mm',
    },
}



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'Nedjelja',
        1: 'Ponedjeljak',
        2: 'Utorak',
        3: 'Srijeda',
        4: 'Četvrtak',
        5: 'Petak',
        6: 'Subota'
    },
    'days_abbrev': {
        0: 'Ne',
        1: 'Po',
        2: 'Ut',
        3: 'Sr',
        4: 'Če',
        5: 'Pe',
        6: 'Su'
    },

    # Months
    'months': {
        1: 'Siječanj',
        2: 'Veljača',
        3: 'Ožujak',
        4: 'Travanj',
        5: 'Svibanj',
        6: 'Lipanj',
        7: 'Srpanj',
        8: 'Kolovoz',
        9: 'Rujan',
        10: 'Listopad',
        11: 'Studeni',
        12: 'Prosinac',
    },
    'months_abbrev': {
        1: 'Sij',
        2: 'Velj',
        3: 'Ožu',
        4: 'Tra',
        5: 'Svi',
        6: 'Lip',
        7: 'Srp',
        8: 'Kol',
        9: 'Ruj',
        10: 'Lis',
        11: 'Stu',
        12: 'Pro',
    },

    # Units of time
    'year': ['{count} godinu', '{count} godine', '{count} godina'],
    'month': ['{count} mjesec', '{count} mjeseca', '{count} mjeseci'],
    'week': ['{count} tjedan', '{count} tjedna', '{count} tjedana'],
    'day': ['{count} dan', '{count} dana', '{count} dana'],
    'hour': ['{count} sat', '{count} sata', '{count} sati'],
    'minute': ['{count} minutu', '{count} minute ', '{count} minuta'],
    'second': ['{count} sekundu', '{count} sekunde', '{count} sekundi'],

    # Relative time
    'ago': 'prije {time}',
    'from_now': 'za {time}',
    'after': 'za {time}',
    'before': 'prije {time}',

    # Ordinals
    'ordinal': '.',

    # Date formats
    'date_formats': {
        'LTS': 'H:mm:ss',
        'LT': 'H:mm',
        'LLLL': 'dddd, D. MMMM YYYY H:mm',
        'LLL': 'D. MMMM YYYY H:mm',
        'LL': 'D. MMMM YYYY',
        'L': 'DD. MM. YYYY',
    },
}



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'Minggu',
        1: 'Senin',
        2: 'Selasa',
        3: 'Rabu',
        4: 'Kamis',
        5: 'Jumat',
        6: 'Sabtu'
    },
    'days_abbrev': {
        0: 'Min',
        1: 'Sen',
        2: 'Sel',
        3: 'Rab',
        4: 'Kam',
        5: 'Jum',
        6: 'Sab'
    },

    # Months
    'months': {
        1: 'Januari',
        2: 'Februari',
        3: 'Maret',
        4: 'April',
        5: 'Mei',
        6: 'Juni',
        7: 'Juli',
        8: 'Agustus',
        9: 'September',
        10: 'Oktober',
        11: 'November',
        12: 'Desember'
    },
    'months_abbrev': {
        1: 'Jan',
        2: 'Feb',
        3: 'Mar',
        4: 'Apr',
        5: 'Mei',
        6: 'Jun',
        7: 'Jul',
        8: 'Ags',
        9: 'Sep',
        10: 'Okt',
        11: 'Nov',
        12: 'Des'
    },

    # Units of time
    'year': '{count} tahun',
    'month': '{count} bulan',
    'week': '{count} minggu',
    'day': '{count} hari',
    'hour': '{count} jam',
    'minute': '{count} menit',
    'second': '{count} detik',

    # Relative time
    'ago': '{time} yang lalu',
    'from_now': '{time} dari sekarang',
    'after': '{time} setelah',
    'before': '{time} sebelum',

    # Date formats
    'date_formats': {
        'LTS': 'HH.mm.ss',
        'LT': 'HH.mm',
        'LLLL': 'dddd, D MMMM YYYY [pukul] HH.mm',
        'LLL': 'D MMMM YYYY [pukul] HH.mm',
        'LL': 'D MMMM YYYY',
        'L': 'DD/MM/YYYY',
    },
}



# -*- coding: utf-8 -*-

translations = {
    # Days
    'days': {
        0: 'dimanĉo',
        1: 'lundo',
        2: 'mardo',
        3: 'merkredo',
        4: 'ĵaŭdo',
        5: 'vendredo',
        6: 'sabato'
    },
    'days_abbrev': {
        0: 'dim',
        1: 'lun',
        2: 'mar',
        3: 'mer',
        4: 'ĵaŭ',
        5: 'ven',
        6: 'sab'
    },

    # Months
    'months': {
        1: 'januaro',
        2: 'februaro',
        3: 'marto',
        4: 'aprilo',
        5: 'majo',
        6: 'junio',
        7: 'julio',
        8: 'aŭgusto',
        9: 'septembro',
        10: 'oktobro',
        11: 'novembro',
        12: 'decembro',
    },
    'months_abbrev': {
        1: 'jan',
        2: 'feb',
        3: 'mar',
        4: 'apr',
        5: 'maj',
        6: 'jun',
        7: 'jul',
        8: 'aŭg',
        9: 'sep',
        10: 'okt',
        11: 'nov',
        12: 'dec',
    },

    # Units of time
    'year': ['{count} jaro', '{count} jaroj'],
    'month': ['{count} monato', '{count} monatoj'],
    'week': ['{count} semajno', '{count} semajnoj'],
    'day': ['{count} tago', '{count} tagoj'],
    'hour': ['{count} horo', '{count} horoj'],
    'minute': ['{count} minuto', '{count} minutoj'],
    'second': ['{count} sekundo', '{count} sekundoj'],

    # Relative time
    'ago': 'antaŭ {time}',
    'from_now': 'je {time}',
    'after': '{time} poste',
    'before': '{time} antaŭe',

    # Meridians
    'meridian': lambda time: 'ATM' if 0 <= time[0] < 12 else 'PTM',

    # Date formats
    'date_formats': {
        'LTS': 'HH:mm:ss',
        'LT': 'HH:mm',
        'LLLL': 'dddd, [la] D[-an de] MMMM, YYYY HH:mm',
        'LLL': 'D[-an de] MMMM, YYYY HH:mm',
        'LL': 'D[-an de] MMMM, YYYY',
        'L': 'YYYY-MM-DD',
    },
}



# -*- coding: utf-8 -*-

from .._compat import decode
from .default import TranslatableMixin


class WordableIntervalMixin(TranslatableMixin):

    def in_words(self, locale=None, separator=' ', _periods=None):
        """
        Get the current interval in words in the current locale.

        Ex: 6 jours 23 heures 58 minutes

        :param locale: The locale to use. Defaults to current locale.
        :type locale: str

        :param separator: The separator to use between each unit
        :type separator: str

        :param _periods: Custom periods to use as word parts
        :type _periods: list or None

        :rtype: str
        """
        if _periods is None:
            _periods = [
                ('week', self.weeks),
                ('day', self.remaining_days),
                ('hour', self.hours),
                ('minute', self.minutes),
                ('second', self.remaining_seconds)
            ]

        parts = []
        for period in _periods:
            unit, count = period
            if abs(count) > 0:
                parts.append(
                    self.translator().transchoice(
                        unit, abs(count), {'count': count}, locale=locale
                    )
                )

        if not parts and abs(self.microseconds) > 0:
            translation = self.translator().transchoice(
                'second', 1,
                {'count': '{:.2f}'.format(abs(self.microseconds) / 1e6)},
                locale=locale
            )
            parts.append(translation)

        return decode(separator.join(parts))

    def __str__(self):
        return self.in_words()

    def __repr__(self):
        return '<{0} [{1}]>'.format(self.__class__.__name__, str(self))
